'''
A generic target capture task
'''
import numpy as np


from riglib.experiment import traits, Sequence, FSMTable, StateTransitions
from riglib.stereo_opengl import ik
from riglib import plants

from riglib.stereo_opengl.window import Window
from .target_graphics import *
import random
import ast

## Plants
# List of possible "plants" that a subject could control either during manual or brain control
cursor = plants.CursorPlant()
shoulder_anchor = np.array([2., 0., -15])
chain_kwargs = dict(link_radii=.6, joint_radii=0.6, joint_colors=(181/256., 116/256., 96/256., 1), link_colors=(181/256., 116/256., 96/256., 1))
chain_20_20_endpt = plants.EndptControlled2LArm(link_lengths=[20, 20], base_loc=shoulder_anchor, **chain_kwargs)
chain_20_20 = plants.RobotArmGen2D(link_lengths=[20, 20], base_loc=shoulder_anchor, **chain_kwargs)

plantlist = dict(
    cursor=cursor,
    chain_20_20=chain_20_20,
    chain_20_20_endpt=chain_20_20_endpt)

class TargetCapture(Sequence):
    '''
    This is a generic cued target capture skeleton, to form as a common ancestor to the most
    common type of motor control task.
    '''
    status = dict(
        wait = dict(start_trial="target"),
        target = dict(enter_target="hold", timeout="timeout_penalty"),
        hold = dict(leave_target="hold_penalty", hold_complete="delay"),
        delay = dict(leave_target="delay_penalty", delay_complete="targ_transition"),
        targ_transition = dict(trial_complete="reward", trial_abort="wait", trial_incomplete="target"),
        timeout_penalty = dict(timeout_penalty_end="targ_transition", end_state=True),
        hold_penalty = dict(hold_penalty_end="targ_transition", end_state=True),
        delay_penalty = dict(delay_penalty_end="targ_transition", end_state=True),
        reward = dict(reward_end="wait", stoppable=False, end_state=True)
    )

    # initial state
    state = "wait"

    target_index = -1 # Helper variable to keep track of which target to display within a trial
    tries = 0 # Helper variable to keep track of the number of failed attempts at a given trial.

    sequence_generators = []

    reward_time = traits.Float(.5, desc="Length of reward dispensation")
    hold_time = traits.Float(.2, desc="Length of hold required at targets before next target appears")
    hold_penalty_time = traits.Float(1, desc="Length of penalty time for target hold error")
    delay_time = traits.Float(0, desc="Length of time after a hold while the next target is on before the go cue")
    delay_penalty_time = traits.Float(1, desc="Length of penalty time for delay error")
    timeout_time = traits.Float(10, desc="Time allowed to go between targets")
    timeout_penalty_time = traits.Float(1, desc="Length of penalty time for timeout error")
    max_attempts = traits.Int(10, desc='The number of attempts of a target chain before skipping to the next one')
    num_targets_per_attempt = traits.Int(2, desc="Minimum number of target acquisitions to be counted as an attempt")

    def init(self):
        self.trial_dtype = np.dtype([('trial', 'u4'), ('index', 'u4'), ('target', 'f8', (3,))])
        super().init()

    def _start_wait(self):
        # Call parent method to draw the next target capture sequence from the generator
        super()._start_wait()

        # number of times this sequence of targets has been attempted
        self.tries = 0

        # index of current target presented to subject
        self.target_index = -1

        # number of targets to be acquired in this trial
        self.chain_length = len(self.targs)

    def _parse_next_trial(self):
        '''Check that the generator has the required data'''
        self.gen_indices, self.targs = self.next_trial
        # TODO error checking
        
        # Update the data sinks with trial information
        self.trial_record['trial'] = self.calc_trial_num()
        for i in range(len(self.gen_indices)):
            self.trial_record['index'] = self.gen_indices[i]
            self.trial_record['target'] = self.targs[i]
            self.sinks.send("trials", self.trial_record)

    def _start_target(self):
        self.target_index += 1

    def _end_target(self):
        '''Nothing generic to do.'''
        pass

    def _start_hold(self):
        '''Nothing generic to do.'''
        pass

    def _while_hold(self):
        '''Nothing generic to do.'''
        pass

    def _end_hold(self):
        '''Nothing generic to do.'''
        pass

    def _start_delay(self):
        '''Nothing generic to do.'''
        pass

    def _while_delay(self):
        '''Nothing generic to do.'''
        pass

    def _end_delay(self):
        '''Nothing generic to do.'''
        pass

    def _start_targ_transition(self):
        '''Nothing generic to do. Child class might show/hide targets'''
        pass

    def _while_targ_transition(self):
        '''Nothing generic to do.'''
        pass

    def _end_targ_transition(self):
        '''Nothing generic to do.'''
        pass

    def _increment_tries(self):
        if self.target_index >= self.num_targets_per_attempt-1:
            self.tries += 1 # only count errors if the minimum number of targets have been acquired
        self.target_index = -1

        if self.tries < self.max_attempts: 
            self.trial_record['trial'] += 1
            for i in range(len(self.gen_indices)):
                self.trial_record['index'] = self.gen_indices[i]
                self.trial_record['target'] = self.targs[i]
                self.sinks.send("trials", self.trial_record)

    def _start_timeout_penalty(self):
        self._increment_tries()

    def _while_timeout_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _end_timeout_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _start_hold_penalty(self):
        self._increment_tries()

    def _while_hold_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _end_hold_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _start_delay_penalty(self):
        self._increment_tries()

    def _while_delay_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _end_delay_penalty(self):
        '''Nothing generic to do.'''
        pass

    def _start_reward(self):
        '''Nothing generic to do.'''
        pass

    def _while_reward(self):
        '''Nothing generic to do.'''
        pass

    def _end_reward(self):
        '''Nothing generic to do.'''
        pass

    ################## State transition test functions ##################
    def _test_start_trial(self, time_in_state):
        '''Start next trial automatically. You may want this to instead be
            - a random delay
            - require some initiation action
        '''
        return True

    def _test_timeout(self, time_in_state):
        return time_in_state > self.timeout_time or self.pause

    def _test_hold_complete(self, time_in_state):
        '''
        Test whether the target is held long enough to declare the
        trial a success

        Possible options
            - Target held for the minimum requred time (implemented here)
            - Sensorized object moved by a certain amount
            - Sensorized object moved to the required location
            - Manually triggered by experimenter
        '''
        return time_in_state > self.hold_time

    def _test_delay_complete(self, time_in_state):
        '''
        Test whether the delay period, when the cursor must stay in place
        while another target is being presented, is over. There should be 
        no delay on the last target in a chain.
        '''
        return self.target_index + 1 == self.chain_length or time_in_state > self.delay_time

    def _test_trial_complete(self, time_in_state):
        '''Test whether all targets in sequence have been acquired'''
        return self.target_index == self.chain_length-1

    def _test_trial_abort(self, time_in_state):
        '''Test whether the target capture sequence should just be skipped due to too many failures'''
        return (not self._test_trial_complete(time_in_state)) and (self.tries==self.max_attempts)

    def _test_trial_incomplete(self, time_in_state):
        '''Test whether the target capture sequence needs to be restarted'''
        return (not self._test_trial_complete(time_in_state)) and (self.tries<self.max_attempts) and not self.pause

    def _test_timeout_penalty_end(self, time_in_state):
        return time_in_state > self.timeout_penalty_time

    def _test_hold_penalty_end(self, time_in_state):
        return time_in_state > self.hold_penalty_time

    def _test_delay_penalty_end(self, time_in_state):
        return time_in_state > self.delay_penalty_time

    def _test_reward_end(self, time_in_state):
        return time_in_state > self.reward_time

    def _test_enter_target(self, time_in_state):
        '''This function is task-specific and not much can be done generically'''
        return False

    def _test_leave_target(self, time_in_state):
        '''This function is task-specific and not much can be done generically'''
        return self.pause

    def update_report_stats(self):
        '''
        see experiment.Experiment.update_report_stats for docs
        '''
        super().update_report_stats()
        self.reportstats['Trial #'] = self.calc_trial_num()
        self.reportstats['Reward/min'] = np.round(self.calc_events_per_min('reward', 120.), decimals=2)

class ScreenTargetCapture(TargetCapture, Window):
    """Concrete implementation of TargetCapture task where targets
    are acquired by "holding" a cursor in an on-screen target"""

    limit2d = traits.Bool(True, desc="Limit cursor movement to 2D")

    sequence_generators = [
        'out_2D', 'centerout_2D', 'centeroutback_2D', 'centerout_2D_select', 'rand_target_chain_2D', 'rand_target_chain_3D', 'corners_2D',
    ]

    hidden_traits = ['cursor_color', 'target_color', 'cursor_bounds', 'cursor_radius', 'plant_hide_rate', 'starting_pos']

    is_bmi_seed = True

    # Runtime settable traits
    target_radius = traits.Float(2, desc="Radius of targets in cm")
    target_color = traits.OptionsList("yellow", *target_colors, desc="Color of the target", bmi3d_input_options=list(target_colors.keys()))
    plant_hide_rate = traits.Float(0.0, desc='If the plant is visible, specifies a percentage of trials where it will be hidden')
    plant_type = traits.OptionsList(*plantlist, bmi3d_input_options=list(plantlist.keys()))
    plant_visible = traits.Bool(True, desc='Specifies whether entire plant is displayed or just endpoint')
    cursor_radius = traits.Float(.5, desc='Radius of cursor in cm')
    cursor_color = traits.OptionsList("dark_purple", *target_colors, desc='Color of cursor endpoint', bmi3d_input_options=list(target_colors.keys()))
    cursor_bounds = traits.Tuple((-10., 10., -10., 10., -10., 10.), desc='(x min, x max, y min, y max, z min, z max)')
    starting_pos = traits.Tuple((5., 0., 5.), desc='Where to initialize the cursor') 

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Initialize the plant
        if not hasattr(self, 'plant'):
            self.plant = plantlist[self.plant_type]
        self.plant.set_bounds(np.array(self.cursor_bounds))
        self.plant.set_color(target_colors[self.cursor_color])
        self.plant.set_cursor_radius(self.cursor_radius)
        self.plant_vis_prev = True
        self.cursor_vis_prev = True

        # Add graphics models for the plant and targets to the window
        if hasattr(self.plant, 'graphics_models'):
            for model in self.plant.graphics_models:
                self.add_model(model)

        # Instantiate the targets
        instantiate_targets = kwargs.pop('instantiate_targets', True)
        if instantiate_targets:

            # 2 targets for delay
            target1 = VirtualCircularTarget(target_radius=self.target_radius, target_color=target_colors[self.target_color])
            target2 = VirtualCircularTarget(target_radius=self.target_radius, target_color=target_colors[self.target_color])

            self.targets = [target1, target2]

        self.target_location = np.array(self.starting_pos).copy()

        # Declare any plant attributes which must be saved to the HDF file at the _cycle rate
        for attr in self.plant.hdf_attrs:
            self.add_dtype(*attr)

    def init(self):
        self.add_dtype('trial', 'u4', (1,))
        self.add_dtype('plant_visible', '?', (1,))
        super().init()
        self.plant.set_endpoint_pos(np.array(self.starting_pos))

    def _cycle(self):
        '''
        Calls any update functions necessary and redraws screen
        '''
        self.move_effector()

        ## Run graphics commands to show/hide the plant if the visibility has changed
        self.update_plant_visibility()
        self.task_data['plant_visible'] = self.plant_visible

        ## Save plant status to HDF file
        plant_data = self.plant.get_data_to_save()
        for key in plant_data:
            self.task_data[key] = plant_data[key]

        # Update the trial index
        self.task_data['trial'] = self.calc_trial_num()

        super()._cycle()

    def move_effector(self):
        '''Move the end effector, if a robot or similar is being controlled'''
        pass

    def run(self):
        '''
        See experiment.Experiment.run for documentation.
        '''
        # Fire up the plant. For virtual/simulation plants, this does little/nothing.
        self.plant.start()
        
        # Include some cleanup in case the parent class has errors
        try:
            super().run()
        finally:
            self.plant.stop()

    ##### HELPER AND UPDATE FUNCTIONS ####
    def update_plant_visibility(self):
        ''' Update plant visibility'''
        if self.plant_visible != self.plant_vis_prev:
            self.plant_vis_prev = self.plant_visible
            self.plant.set_visibility(self.plant_visible)

    #### TEST FUNCTIONS ####
    def _test_enter_target(self, ts):
        '''
        return true if the distance between center of cursor and target is smaller than the cursor radius
        '''
        cursor_pos = self.plant.get_endpoint_pos()
        d = np.linalg.norm(cursor_pos - self.targs[self.target_index])
        return d <= (self.target_radius - self.cursor_radius) or super()._test_enter_target(ts)

    def _test_leave_target(self, ts):
        '''
        return true if cursor moves outside the exit radius
        '''
        cursor_pos = self.plant.get_endpoint_pos()
        d = np.linalg.norm(cursor_pos - self.targs[self.target_index])
        rad = self.target_radius - self.cursor_radius
        return d > rad or super()._test_leave_target(ts)

    #### STATE FUNCTIONS ####
    def _start_wait(self):
        super()._start_wait()

        if self.calc_trial_num() == 0:

            # Instantiate the targets here so they don't show up in any states that might come before "wait"
            for target in self.targets:
                for model in target.graphics_models:
                    self.add_model(model)
                    target.hide()

    def _start_target(self):
        super()._start_target()

        # Show target if it is hidden (this is the first target, or previous state was a penalty)
        target = self.targets[self.target_index % 2]
        if self.target_index == 0:
            target.move_to_position(self.targs[self.target_index])
            target.show()
            self.sync_event('TARGET_ON', self.gen_indices[self.target_index])
        self.target_location = self.targs[self.target_index] # save for BMILoop

    def _start_hold(self):
        super()._start_hold()
        self.sync_event('CURSOR_ENTER_TARGET', self.gen_indices[self.target_index])

    def _start_delay(self):
        super()._start_delay()

        # Make next target visible unless this is the final target in the trial
        next_idx = (self.target_index + 1)
        if next_idx < self.chain_length:
            target = self.targets[next_idx % 2]
            target.move_to_position(self.targs[next_idx])
            target.show()
            self.sync_event('TARGET_ON', self.gen_indices[next_idx])
        else:
            # This delay state should only last 1 cycle, don't sync anything
            pass

    def _start_targ_transition(self):
        super()._start_targ_transition()
        if self.target_index == -1:

            # Came from a penalty state
            pass
        elif self.target_index + 1 < self.chain_length:

            # Hide the current target if there are more
            self.targets[self.target_index % 2].hide()
            self.sync_event('TARGET_OFF', self.gen_indices[self.target_index])

    def _start_hold_penalty(self):
        self.sync_event('HOLD_PENALTY') 
        super()._start_hold_penalty()
        # Hide targets
        for target in self.targets:
            target.hide()
            target.reset()

    def _end_hold_penalty(self):
        super()._end_hold_penalty()
        self.sync_event('TRIAL_END')

    def _start_delay_penalty(self):
        self.sync_event('DELAY_PENALTY') 
        super()._start_delay_penalty()
        # Hide targets
        for target in self.targets:
            target.hide()
            target.reset()

    def _end_delay_penalty(self):
        super()._end_delay_penalty()
        self.sync_event('TRIAL_END')
        
    def _start_timeout_penalty(self):
        self.sync_event('TIMEOUT_PENALTY')
        super()._start_timeout_penalty()
        # Hide targets
        for target in self.targets:
            target.hide()
            target.reset()

    def _end_timeout_penalty(self):
        super()._end_timeout_penalty()
        self.sync_event('TRIAL_END')

    def _start_reward(self):
        self.targets[self.target_index % 2].cue_trial_end_success()
        self.sync_event('REWARD')
        
    def _end_reward(self):
        super()._end_reward()
        self.sync_event('TRIAL_END')

        # Hide targets
        for target in self.targets:
            target.hide()
            target.reset()
            
    #### Generator functions ####
    '''
    Note to self: because of the way these get into the database, the parameters don't
    have human-readable descriptions like the other traits. So it is useful to define
    the descriptions elsewhere, in models.py under Generator.to_json().

    Ideally someone should take the time to reimplement generators as their own classes
    rather than static methods that belong to a task.
    '''
    @staticmethod
    def static(pos=(0,0,0), ntrials=0):
        '''Single location, finite (ntrials!=0) or infinite (ntrials==0)'''
        if ntrials == 0:
            while True:
                yield [0], np.array(pos)
        else:
            for _ in range(ntrials):
                yield [0], np.array(pos)

    @staticmethod
    def out_2D(nblocks=100, ntargets=8, distance=10, origin=(0,0,0)):
        '''
        Generates a sequence of 2D (x and z) targets at a given distance from the origin

        Parameters
        ----------
        nblocks : int
            The number of ntarget pairs in the sequence.
        ntargets : int
            The number of equally spaced targets
        distance : float
            The distance in cm between the center and peripheral targets.
        origin : 3-tuple
            Location of the central targets around which the peripheral targets span

        Returns
        -------
        [nblocks*ntargets x 1] array of tuples containing trial indices and [1 x 3] target coordinates

        '''
        rng = np.random.default_rng()
        for _ in range(nblocks):
            order = np.arange(ntargets) + 1 # target indices, starting from 1
            rng.shuffle(order)
            for t in range(ntargets):
                idx = order[t]
                theta = 2*np.pi*(3-idx)/ntargets # put idx 1 at 12 o'clock
                pos = np.array([
                    distance*np.cos(theta),
                    0,
                    distance*np.sin(theta)
                ]).T
                yield [idx], [pos + origin]

    @staticmethod
    def centerout_2D(nblocks=100, ntargets=8, distance=10, origin=(0,0,0)):
        '''
        Pairs of central targets at the origin and peripheral targets centered around the origin

        Returns
        -------
        [nblocks*ntargets x 1] array of tuples containing trial indices and [2 x 3] target coordinates
        '''
        gen = ScreenTargetCapture.out_2D(nblocks, ntargets, distance, origin)
        for _ in range(nblocks*ntargets):
            idx, pos = next(gen)
            targs = np.zeros([2, 3]) + origin
            targs[1,:] = pos[0]
            indices = np.zeros([2,1])
            indices[1] = idx
            yield indices, targs

    @staticmethod
    def centerout_2D_select(nblocks=100, ntargets=8, distance=10, origin=(0,0,0), target_idx=[]):
        '''
        Generates a sequence of 2D (x and z) targets at a given distance from the origin, 
        but lets you select which targets out of the total number you want to keep
        '''
        if type(target_idx) not in (list, tuple) or len(target_idx) == 0:
            raise ValueError(f'Malformed target_idx selection "{target_idx}"! Input a list of [idx1, idx2, ...]')
        gen = ScreenTargetCapture.centerout_2D(nblocks, ntargets, distance, origin)
        while True:
            try:
                next_indices, next_targs = next(gen)
                if next_indices[1] in target_idx:
                    yield next_indices, next_targs # only yield the targets that match target_idx
            except StopIteration:
                break

    @staticmethod
    def centeroutback_2D(nblocks=100, ntargets=8, distance=10, origin=(0,0,0)):
        '''
        Triplets of central targets, peripheral targets, and central targets

        Returns
        -------
        [nblocks*ntargets x 1] array of tuples containing trial indices and [3 x 3] target coordinates
        '''
        gen = ScreenTargetCapture.out_2D(nblocks, ntargets, distance, origin)
        for _ in range(nblocks*ntargets):
            idx, pos = next(gen)
            targs = np.zeros([3, 3]) + origin
            targs[1,:] = pos[0]
            indices = np.zeros([3,1])
            indices[1] = idx
            yield indices, targs
    
    @staticmethod
    def rand_target_chain_2D(ntrials=100, chain_length=1, boundaries=(-12,12,-12,12)):
        '''
        Generates a sequence of 2D (x and z) target pairs.

        Parameters
        ----------
        ntrials : int
            The number of target chains in the sequence.
        chain_length : int
            The number of targets in each chain
        boundaries: 4 element Tuple
            The limits of the allowed target locations (-x, x, -z, z)

        Returns
        -------
        [ntrials x chain_length x 3] array of target coordinates
        '''
        rng = np.random.default_rng()
        idx = 0
        for t in range(ntrials):

            # Choose a random sequence of points within the boundaries
            pts = rng.uniform(size=(chain_length, 3))*((boundaries[1]-boundaries[0]),
                0, (boundaries[3]-boundaries[2]))
            pts = pts+(boundaries[0], 0, boundaries[2])
            yield idx+np.arange(chain_length), pts
            idx += chain_length
    
    @staticmethod
    def rand_target_chain_3D(ntrials=100, chain_length=1, boundaries=(-12,12,-10,10,-12,12)):
        '''
        Generates a sequence of 3D target pairs.
        Parameters
        ----------
        ntrials : int
            The number of target chains in the sequence.
        chain_length : int
            The number of targets in each chain
        boundaries: 6 element Tuple
            The limits of the allowed target locations (-x, x, -y, y, -z, z)

        Returns
        -------
        [ntrials x chain_length x 3] array of target coordinates
        '''
        rng = np.random.default_rng()
        idx = 0
        for t in range(ntrials):

            # Choose a random sequence of points within the boundaries
            pts = rng.uniform(size=(chain_length, 3))*((boundaries[1]-boundaries[0]),
                (boundaries[3]-boundaries[2]), (boundaries[5]-boundaries[4]))
            pts = pts+(boundaries[0], boundaries[2], boundaries[4])
            yield idx+np.arange(chain_length), pts
            idx += chain_length

    @staticmethod
    def corners_2D(nblocks=5, chain_length=1, corners=(-8,8,-8,8)):
        '''
        Generates a sequence of 2D (x and z) targets at the given 4 corners

        Parameters
        ----------
        nblocks : 3-tuple
            Number of blocks
        chain_length : int
            The number of targets in each chain before a reward is given
        corners : 4-tuple
            Location of the edges of the screen (-x, x, -y, y)

        Returns
        -------
        [nblocks*4 x 1] array of tuples containing trial indices and [1 x 3] target coordinates

        '''
        ntargets = 4
        corners = np.array([
            [corners[0], 0, corners[2]],
            [corners[0], 0, corners[3]],
            [corners[1], 0, corners[2]],
            [corners[1], 0, corners[3]]
        ])
        target_order = []
        rng = np.random.default_rng()
        for _ in range(nblocks):
            order = np.arange(ntargets) + 1 # target indices, starting from 1
            rng.shuffle(order)
            target_order = np.concatenate((target_order, order), axis=0)

        # Spit out trials in groups of chain_length
        ntrials = nblocks*4//chain_length
        for t in range(ntrials):
            idx = target_order[int(t*chain_length):int(t*chain_length+chain_length)]
            pos = [corners[int(i-1),:] for i in idx]
            yield idx, pos

class ScreenReachAngle(ScreenTargetCapture):
    '''
    A modified task that requires the cursor to move in the right direction towards the target, 
    without actually needing to arrive at the target. If the maximum angle is exceeded, a reach 
    penalty is applied. No hold or delay period.

    Only works for sequences with 1 target in a chain. 
    '''

    status = dict(
        wait = dict(start_trial="target"),
        target = dict(reach_success="targ_transition", timeout="timeout_penalty", leave_bounds="reach_penalty"),
        targ_transition = dict(trial_complete="reward", trial_abort="wait", trial_incomplete="target"),
        timeout_penalty = dict(timeout_penalty_end="targ_transition", end_state=True),
        reach_penalty = dict(reach_penalty_end="targ_transition", end_state=True),
        reward = dict(reward_end="wait", stoppable=False, end_state=True)
    )

    sequence_generators = [
        'out_2D', 'rand_target_chain_2D', 'rand_target_chain_3D', 'discrete_targets_2D',
    ]

    max_reach_angle = traits.Float(90., desc="Angle defining the boundaries between the starting position of the cursor and the target")
    reach_penalty_time = traits.Float(1, desc="Length of penalty time for target hold error")
    reach_fraction = traits.Float(0.5, desc="Fraction of the distance between the reach start and the target before a reward")
    start_radius = traits.Float(1., desc="Buffer around reach start allowed in bounds (cm)")

    exclude_parent_traits = ['hold_time', 'hold_penalty_time', 'delay_time', 'delay_penalty_time']

    def _start_target(self):
        super()._start_target()

        # Define a reach start and reach target position whenever the target appears
        self.reach_start = self.plant.get_endpoint_pos().copy()
        self.reach_target = self.targs[self.target_index]

    def _test_leave_bounds(self, ts):
        '''
        Check whether the cursor is in the boundary defined by reach_start, target_pos,
        and max_reach_angle.
        '''

        # Calculate the angle between the vectors from the start pos to the current cursor and target
        a = self.plant.get_endpoint_pos() - self.reach_start
        b = self.reach_target - self.reach_start
        cursor_target_angle = np.arccos(np.dot(a, b)/np.linalg.norm(a)/np.linalg.norm(b))

        # If that angle is more than half the maximum, we are outside the bounds
        out_of_bounds = np.degrees(cursor_target_angle) > self.max_reach_angle / 2

        # But also allow a target radius around the reach_start 
        away_from_start = np.linalg.norm(self.plant.get_endpoint_pos() - self.reach_start) > self.start_radius

        return away_from_start and out_of_bounds

    def _test_reach_success(self, ts):
        dist_traveled = np.linalg.norm(self.plant.get_endpoint_pos() - self.reach_start)
        dist_total = np.linalg.norm(self.reach_target - self.reach_start)
        dist_total -= (self.target_radius - self.cursor_radius)
        return dist_traveled/dist_total > self.reach_fraction

    def _start_reach_penalty(self):
        self.sync_event('OTHER_PENALTY')
        self._increment_tries()
        
        # Hide targets
        for target in self.targets:
            target.hide()
            target.reset()

    def _end_reach_penalty(self):
        self.sync_event('TRIAL_END')

    def _test_reach_penalty_end(self, ts):
        return ts > self.reach_penalty_time

    @staticmethod
    def discrete_targets_2D(nblocks=100, ntargets=3, boundaries=(-6,6,-3,3)):
        '''
        Generates a sequence of 2D (x and z) target pairs that don't overlap

        Parameters
        ----------
        nblocks : int
            The number of ntarget pairs in the sequence.
        ntargets : int
            The number of unique targets (up to 9 maximum)
        boundaries: 4 element Tuple
            The limits of the allowed target locations (-x, x, -z, z)

        Returns
        -------
        [ntrials x ntargets x 3] array of target coordinates
        '''
        targets = np.array([
            [0, 0.5],
            [1, 0.5],
            [1, 0],
            [0, 0],
            [0.25, 0.25],
            [0.75, 0.25],
            [0.25, 0.75],
            [0.75, 0.75],
            [0.5, 1],
        ])
        rng = np.random.default_rng()
        for _ in range(nblocks):
            order = np.arange(ntargets) # target indices
            rng.shuffle(order)
            for t in range(ntargets):
                idx = order[t]
                pts = targets[idx]*((boundaries[1]-boundaries[0]),
                    (boundaries[3]-boundaries[2]))
                pts = pts+(boundaries[0], boundaries[2])
                pos = np.array([pts[0], 0, pts[1]])
                yield [idx], [pos]


class SequenceCapture(ScreenTargetCapture):

    '''
    This is a sequence task in which a 2nd target appears after subjects acquire the 1st target, and
    a 3rd target (additional target) appears while they are moving the cursor to the 2nd target.
    '''

    status = dict(
        wait = dict(start_trial="target"),
        target = dict(enter_target="hold", timeout="timeout_penalty", show_additional_target='additional_target'),
        additional_target = dict(enter_target="hold", timeout="timeout_penalty", enter_incorrect_target="incorrect_target_penalty"),
        hold = dict(leave_target="hold_penalty", hold_complete="delay"),
        delay = dict(leave_target="delay_penalty", delay_complete="targ_transition"),
        targ_transition = dict(trial_complete="reward", trial_abort="wait", trial_incomplete="target"),
        timeout_penalty = dict(timeout_penalty_end="targ_transition", end_state=True),
        hold_penalty = dict(hold_penalty_end="targ_transition", end_state=True),
        delay_penalty = dict(delay_penalty_end="targ_transition", end_state=True),
        incorrect_target_penalty = dict(incorrect_target_penalty_end="targ_transition", end_state=True),
        reward = dict(reward_end="wait", stoppable=False, end_state=True)
    )

    sequence_generators = ['out_2D_sequence','sequence_2D', 'centerout_2D_different_center']
    incorrect_target_penalty_time = traits.Float(1, desc="Length of penalty time for acquiring an incorrect target")
    random_target_appearance_dist = traits.Tuple((3.5, 3.5), desc="Another target appear when the cursor passed a certain distance from the 1st target in cm")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Instantiate the targets
        instantiate_targets = kwargs.pop('instantiate_targets', True)
        if instantiate_targets:

            # Need three targets for sequence task
            target1 = VirtualCircularTarget(target_radius=self.target_radius, target_color=target_colors[self.target_color])
            target2 = VirtualCircularTarget(target_radius=self.target_radius, target_color=target_colors[self.target_color])
            target3 = VirtualCircularTarget(target_radius=self.target_radius, target_color=target_colors[self.target_color])

            self.targets = [target1, target2, target3]

    #### STATE FUNCTIONS ####

    def _start_wait(self):
        '''
        Target apperance distance is sampled from the uniform distribution at the beginning of wait state.
        '''
        super()._start_wait()
        s, e = self.random_target_appearance_dist
        self.target_appearance_distance = random.uniform(s,e)

    def _start_delay(self):
        next_idx = (self.target_index + 1)
        if next_idx == 1 and next_idx < self.chain_length: # Delay period is only for the 2nd target
            target = self.targets[next_idx % self.chain_length]
            target.move_to_position(self.targs[next_idx])
            target.show()
            self.sync_event('TARGET_ON', self.gen_indices[next_idx])     
        else:
            # Don't sync anything when the 1st target and 3rd target appear
            pass

    def _start_targ_transition(self):
        if self.target_index == -1:

            # Came from a penalty state
            pass
        elif self.target_index + 1 < self.chain_length:

            # Hide the current target if there are more
            self.targets[self.target_index % self.chain_length].hide()
            self.sync_event('TARGET_OFF', self.gen_indices[self.target_index])

    def _start_additional_target(self):
        next_idx = (self.target_index + 1)
        target = self.targets[next_idx % self.chain_length]
        target.move_to_position(self.targs[next_idx])
        target.show()
        self.sync_event('TARGET_ON', self.gen_indices[next_idx])    

    def _start_incorrect_target_penalty(self):
        self._increment_tries()
        self.sync_event('OTHER_PENALTY') 
        # Hide targets
        for target in self.targets:
            target.hide()
            target.reset()

    def _end_incorrect_target_penalty(self):
        self.sync_event('TRIAL_END') 

    def _start_reward(self):
        self.targets[self.target_index % self.chain_length].cue_trial_end_success()
        self.sync_event('REWARD')

    #### TEST FUNCTIONS ####

    def _test_hold_complete(self, time_in_state):
        '''
        Test whether the target is held long enough to declare the
        trial a success

        Possible options
            - Target held for the minimum requred time (implemented here)
            - Sensorized object moved by a certain amount
            - Sensorized object moved to the required location
            - Manually triggered by experimenter
        '''
        if self.target_index == 0: # the 1st target
            hold_state = time_in_state > self.hold_time
        elif self.target_index == self.chain_length-1: # the last target
            hold_state = time_in_state > self.hold_time
        else:
            hold_state = True
        return hold_state

    def _test_delay_complete(self, time_in_state):
        '''
        Test whether the delay period, when the cursor must stay in place
        while another target is being presented, is over. There should be 
        no delay on the last target in a chain.
        '''
        if self.target_index == 0:
            delay_state = time_in_state > self.delay_time
        else:
            delay_state = True # No delay period for the 3rd target
        return delay_state

    def _test_enter_incorrect_target(self, ts):
        cursor_pos = self.plant.get_endpoint_pos()
        d = np.linalg.norm(cursor_pos - self.targs[self.target_index+1]) # Measure distance between the cursor and the next target
        rad = self.target_radius - self.cursor_radius

        return d < rad   

    def _test_incorrect_target_penalty_end(self, time_in_state):
        return time_in_state > self.incorrect_target_penalty_time

    def _test_show_additional_target(self,ts):
        cursor_pos = self.plant.get_endpoint_pos()
        d = np.linalg.norm(cursor_pos - self.targs[self.target_index-1]) # Measure distance between the cursor and the previous target
        rad = self.target_radius - self.cursor_radius

        if self.chain_length < 3: # The 3rd targets isn't shown if chain_length is less than 3 (ex. standard center out task)
           additional_target_state = False
        elif self.target_index == 1: # Target_index == 1 is when the subject is moving the cursor to the 2nd target 
            additional_target_state = d > self.target_appearance_distance
        else:
            additional_target_state = False
        return additional_target_state


    #### Generator functions ####
    @staticmethod
    def out_2D_sequence(nblocks=100, distance=6.8):
        '''
        Generates a sequence of 2D (x and z) targets at a given distance from the origin
        The center target positions change depnding on the peripheral target at each trial
        The number of targets are fixed to 8.

        Parameters
        ----------
        nblocks : int
            The number of ntarget pairs in the sequence.
        distance : float
            The distance in cm between the center and peripheral targets.

        Returns
        -------
        [nblocks*ntargets x 1] array of tuples containing trial indices and [1 x 3] target coordinates

        '''
        ntargets = 8
        rng = np.random.default_rng()
        for _ in range(nblocks):
            order = np.arange(ntargets) + 1 # target indices, starting from 1
            rng.shuffle(order)
            for t in range(ntargets):
                idx = order[t]
                theta = 2*np.pi*(1-idx)/4 + np.pi/2 # put idx 1 at 12 o'clock
                print(theta)
                if idx <= 4:
                    pos = np.array([distance*np.cos(theta)+distance/2,0,distance*np.sin(theta)]).T
                elif idx >=5:
                    pos = np.array([distance*np.cos(theta)-distance/2,0,distance*np.sin(theta)]).T
                yield [idx], [pos]

    @staticmethod
    def centerout_2D_different_center(nblocks=100, distance=6.8):
        '''
        Pairs of a center target and a peripheral target. 
        The center target position changes depending on the periphetal target at each trial.
        The number of targets is fixed to 8.

        Returns
        -------
        [nblocks*ntargets x 1] array of tuples containing trial indices and [2 x 3] target coordinates
        '''
        ntargets = 8
        gen = SequenceCapture.out_2D_sequence(nblocks=nblocks, distance=distance)
        for _ in range(nblocks*ntargets):
            idx, pos = next(gen)
            targs = np.zeros([2, 3])
            if idx[0] <= 4:
                targs[0,:] = np.array([distance/2,0,0])
            else:
                targs[0,:] = np.array([-distance/2,0,0])
            targs[1,:] = pos[0]
            indices = np.zeros([2,1])
            indices[1] = idx
            yield indices, targs

    @staticmethod
    def sequence_2D(nblocks=100, distance=6.8):
        '''
        Pairs of the 1st, 2nd, and 3rd target for the sequence task.

        Returns
        -------
        [nblocks*ntargets x 1] array of tuples containing trial indices and [3 x 3] target coordinates
        '''
        ntargets = 8
        gen = SequenceCapture.out_2D_sequence(nblocks=nblocks, distance=distance)
        for _ in range(nblocks*ntargets):
            idx, pos = next(gen)
            targs = np.zeros([3, 3])
            indices = np.zeros([3,1])
            idx = idx[0]
            pos = pos[0]

            if idx >= 5:
                targs[0,:] = np.array([distance/2,0,0]).T
                indices[0] = 0 # The target for trial initiation is 0

                targs[1,:] = np.array([-distance/2,0,0]).T
                indices[1] = 4

                targs[2,:] = pos
                indices[2] = idx
            
            elif idx <= 4:
                targs[0,:] = np.array([-distance/2,0,0]).T
                indices[0] = 0 #The target for trial initiation is 0

                targs[1,:] = np.array([distance/2,0,0]).T
                indices[1] = 6

                targs[2,:] = pos            
                indices[2] = idx

            yield indices, targs


class HandConstrainedEyeCapture(ScreenTargetCapture):
    '''
    Saccade task with holding another target with hand. Subjects need to hold an initial target with their hand. 
    Then they need to fixate the first eye target and make a saccade for the second eye target 
    '''

    fixation_radius = traits.Float(2.5, desc="Distance from center that is considered a broken fixation")
    fixation_penalty_time = traits.Float(1.0, desc="Time in fixation penalty state")
    fixation_target_color = traits.OptionsList("cyan", *target_colors, desc="Color of the eye target under fixation state", bmi3d_input_options=list(target_colors.keys()))
    eye_target_color = traits.OptionsList("white", *target_colors, desc="Color of the eye target", bmi3d_input_options=list(target_colors.keys()))
    fixation_radius_buffer = traits.Float(.5, desc="additional radius for eye target")

    status = dict(
        wait = dict(start_trial="target"),
        target = dict(leave_target2="hold_penalty",timeout="timeout_penalty",enter_target="hold"),
        hold = dict(leave_target2="hold_penalty",leave_target="target", gaze_target="fixation"), # must hold an initial hand-target and eye-target
        fixation = dict(leave_target="delay_penalty",hold_complete="delay", fixation_break="fixation_penalty"), # must hold an initial hand-target and eye-target to initiate a trial
        delay = dict(leave_target="delay_penalty", delay_complete="targ_transition", fixation_break="fixation_penalty"),
        targ_transition = dict(trial_complete="reward", trial_abort="wait", trial_incomplete="target"),
        timeout_penalty = dict(timeout_penalty_end="targ_transition", end_state=True),
        hold_penalty = dict(hold_penalty_end="targ_transition", end_state=True),
        delay_penalty = dict(delay_penalty_end="targ_transition", end_state=True),
        fixation_penalty = dict(fixation_penalty_end="targ_transition",end_state=True),
        reward = dict(reward_end="wait", stoppable=False, end_state=True)
    )
 
    sequence_generators = ['row_target','sac_hand_2d']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Instantiate the targets
        instantiate_targets = kwargs.pop('instantiate_targets', True)
        if instantiate_targets:

            # Target 1 and 2 are for saccade. Target 3 is for hand
            target1 = VirtualCircularTarget(target_radius=self.fixation_radius, target_color=target_colors[self.eye_target_color])
            target2 = VirtualCircularTarget(target_radius=self.fixation_radius, target_color=target_colors[self.eye_target_color])
            target3 = VirtualCircularTarget(target_radius=self.target_radius, target_color=target_colors[self.target_color])

            self.targets = [target1, target2]
            self.targets_hand = [target3]
        
    def _parse_next_trial(self):
        '''Check that the generator has the required data'''
        self.gen_indices, self.targs = self.next_trial # 2 target positions for hand and eye

        # Update the data sinks with trial information
        self.trial_record['trial'] = self.calc_trial_num() # TODO save both eye and hand target positions
        for i in range(len(self.gen_indices)):
            self.trial_record['index'] = self.gen_indices[i]
            self.trial_record['target'] = self.targs[i]
            self.sinks.send("trials", self.trial_record)

    def _test_gaze_target(self,ts):
        '''
        Check whether eye positions from a target are within the fixation distance
        '''
        # Distance of an eye position from a target position
        eye_pos = self.calibrated_eye_pos
        target_pos = np.delete(self.targs[self.target_index],1)
        d_eye = np.linalg.norm(eye_pos - target_pos)
        return (d_eye <= self.fixation_radius + self.fixation_radius_buffer) or self.pause
        
    def _test_fixation_break(self,ts):
        '''
        Triggers the fixation_penalty state when eye positions are outside fixation distance
        '''
        # Distance of an eye position from a target position
        eye_pos = self.calibrated_eye_pos
        target_pos = np.delete(self.targs[self.target_index],1)
        d_eye = np.linalg.norm(eye_pos - target_pos)
        return (d_eye > self.fixation_radius + self.fixation_radius_buffer) or self.pause
    
    def _test_fixation_penalty_end(self,ts):
        return (ts > self.fixation_penalty_time)
    
    def _test_enter_target(self, ts):
        '''
        return true if the distance between center of cursor and target is smaller than the cursor radius
        '''
        cursor_pos = self.plant.get_endpoint_pos()
        d = np.linalg.norm(cursor_pos - self.targs[-1]) # hand must be within the initial target
        return d <= (self.target_radius - self.cursor_radius) or self.pause

    def _test_leave_target(self, ts):
        '''
        return true if cursor moves outside the exit radius
        '''
        cursor_pos = self.plant.get_endpoint_pos()
        d = np.linalg.norm(cursor_pos - self.targs[-1]) # hand must be within the initial target
        return d > (self.target_radius - self.cursor_radius) or self.pause

    def _test_leave_target2(self, ts):
        '''
        return true if cursor moves outside the exit radius (This is for the second target state)
        '''
        if self.target_index > 0:
            cursor_pos = self.plant.get_endpoint_pos()
            d = np.linalg.norm(cursor_pos - self.targs[-1]) # hand must be within the initial target
            return d > (self.target_radius - self.cursor_radius) or self.pause
    
    def _start_wait(self):
        super()._start_wait()
        # Redefine chain length because targs in this task has both eye and hand targets
        self.chain_length = len(self.targets)

        # Initialize fixation state
        self.num_hold_state = 0

        if self.calc_trial_num() == 0:

            # Instantiate the targets here so they don't show up in any states that might come before "wait"
            for target in self.targets:
                for model in target.graphics_models:
                    self.add_model(model)
                    target.hide()

            for target in self.targets_hand:
                for model in target.graphics_models:
                    self.add_model(model)
                    target.hide()

    def _start_target(self):
        if self.num_hold_state == 0:
            self.target_index += 1 # target index shouldn't be incremented after hold break loop

            # Show target if it is hidden (this is the first target, or previous state was a penalty)
            target_hand = self.targets_hand[0]
            if self.target_index == 0:
                target_hand.move_to_position(self.targs[-1])
                target_hand.show()
                self.sync_event('TARGET_ON', self.gen_indices[-1])
                
        else:
            target = self.targets[self.target_index % 2]
            target.hide() # hide hand target
            self.sync_event('EYE_TARGET_OFF', self.gen_indices[self.target_index % 2])

    def _start_hold(self):
        #self.sync_event('CURSOR_ENTER_TARGET', self.gen_indices[self.target_index])
        self.num_hold_state = 1

        # Show target if it is hidden (this is the first target, or previous state was a penalty)
        target = self.targets[self.target_index % 2]
        if self.target_index == 0:
            target.move_to_position(self.targs[self.target_index])
            target.show()
            self.sync_event('EYE_TARGET_ON', self.gen_indices[self.target_index])

    def _start_fixation(self):
        self.num_hold_state = 0
        self.targets[self.target_index].sphere.color = target_colors[self.fixation_target_color] # change target color in fixation state
        self.sync_event('FIXATION', self.gen_indices[self.target_index])

    def _while_fixation(self):
        pass

    def _end_fixation(self):
        pass

    def _start_delay(self):
        # Make next target visible unless this is the final target in the trial
        next_idx = (self.target_index + 1)
        if next_idx < self.chain_length:
            target = self.targets[next_idx % 2]
            target.move_to_position(self.targs[next_idx % 2])
            target.show()
            self.sync_event('EYE_TARGET_ON', self.gen_indices[next_idx % 2])
        else:
            # This delay state should only last 1 cycle, don't sync anything
            pass

    def _start_targ_transition(self):
        if self.target_index == -1:

            # Came from a penalty state
            pass
        elif self.target_index + 1 < self.chain_length:

            # Hide the current target if there are more
            self.targets[self.target_index % 2].hide()
            self.sync_event('EYE_TARGET_OFF', self.gen_indices[self.target_index])

    def _start_timeout_penalty(self):
        super()._start_timeout_penalty()
        self.num_hold_state = 0
        for target in self.targets_hand:
            target.hide()
            target.reset()
            
    def _start_hold_penalty(self):
        super()._start_hold_penalty()
        self.num_hold_state = 0
        # Hide targets
        for target in self.targets_hand:
            target.hide()
            target.reset()

    def _start_delay_penalty(self):
        super()._start_delay_penalty()
        # Hide targets
        for target in self.targets_hand:
            target.hide()
            target.reset()

    def _start_fixation_penalty(self):
        self._increment_tries()
        self.sync_event('FIXATION_PENALTY') 

        # Hide targets
        for target in self.targets:
            target.hide()
            target.reset()

        for target in self.targets_hand:
            target.hide()
            target.reset()

    def _end_fixation_penalty(self):
        self.sync_event('TRIAL_END')

    def _start_reward(self):
        super()._start_reward()
        # Hide targets
        for target in self.targets_hand:
            target.hide()
            target.reset()        

    # Generator functions
    @staticmethod
    def row_target(nblocks=20, ntargets=3, dx=5.,offset1=(0,0,-2),offset2=(0,0,6.),offset3=(0,0,-7.5),origin=(0,0,0),seed=0):
        '''
        Generates a sequence of 3D for 2 eye targets and 1 hand target at a given distance from the origin

        Parameters
        ----------
        nblocks : int
            The number of ntarget pairs in the sequence.
        ntargets : int
            The number of equally spaced targets
        distance : float
            The distance in cm between targets
        offset1 : 3-tuple
            y location of the first eye target
        offset2 : 3-tuple
            y location of the second eye target
        offset3 : 3-tuple
            y location of the hand target
        origin : 3-tuple
            Location of the central targets

        Returns
        -------
        [nblocks*ntargets x 1] array of tuples containing trial indices and [1 x 3] target coordinates

        '''
        rng = np.random.default_rng(seed=seed)
        for _ in range(nblocks):
            order = np.arange(ntargets**3)
            rng.shuffle(order)
            x_pos_candidate = [-dx,0,dx]
            for t in range(ntargets**3):
                idx = np.base_repr(order[t],3).zfill(3) # convert a decimal number to ternary

                # Target index for hand target, initial eye target, final eye target
                idx1 = int(idx[0])
                idx2 = int(idx[1])
                idx3 = int(idx[2])

                # Get positions for each target
                x_pos1 = x_pos_candidate[idx1]
                x_pos2 = x_pos_candidate[idx2]
                x_pos3 = x_pos_candidate[idx3]
                pos1 = np.array([x_pos1,0,0]).T
                pos2 = np.array([x_pos2,0,0]).T
                pos3 = np.array([x_pos3,0,0]).T

                yield [idx1],[idx2],[idx3],[pos1+offset1+origin],[pos2+offset2+origin],[pos3+offset3+origin]

    @staticmethod
    def sac_hand_2d(nblocks=20, ntargets=3, dx=10,offset1=(0,0,-2),offset2=(0,0,6.),offset3=(0,0,-7.5),origin=(0,0,0),seed=0):
        '''
        Pairs of hand targets and eye targets

        Returns
        -------
        [nblocks*ntargets x 1] array of tuples containing trial indices and [3 x 3] target coordinates
        '''

        gen = HandConstrainedEyeCapture.row_target(nblocks=nblocks,ntargets=ntargets,dx=dx,offset1=offset1,offset2=offset2,offset3=offset3,origin=origin,seed=seed)
        for _ in range(nblocks*(ntargets**3)):
            idx1,idx2,idx3,pos1,pos2,pos3 = next(gen)

            targs = np.zeros([3, 3])
            targs[0,:] = pos1[0]
            targs[1,:] = pos2[0]
            targs[2,:] = pos3[0]

            indices = np.zeros([3,1])
            indices[0] = idx1[0]
            indices[1] = idx2[0] + ntargets
            indices[2] = idx3[0]

            yield indices, targs


class ScreenTargetCapture_Saccade(ScreenTargetCapture):
    '''
    Center-out saccade task. The controller for the cursor position is eye position.
    Hand cursor is also visible. You should remove the hand cursor by setting cursor_radius to 0 as needed.
    '''

    fixation_radius_buffer = traits.Float(.5, desc="additional radius for eye target")
    target_color = traits.OptionsList("white", *target_colors, desc="Color of the target", bmi3d_input_options=list(target_colors.keys()))
    fixation_target_color = traits.OptionsList("cyan", *target_colors, desc="Color of the eye target under fixation state", bmi3d_input_options=list(target_colors.keys()))

    def _test_enter_target(self, ts):
        '''
        Check whether eye positions from a target are within the fixation distance
        '''
        # Distance of an eye position from a target position
        eye_pos = self.calibrated_eye_pos
        target_pos = np.delete(self.targs[self.target_index],1)
        d_eye = np.linalg.norm(eye_pos - target_pos)
        return (d_eye <= self.target_radius + self.fixation_radius_buffer) or self.pause

    def _test_leave_target(self, ts):
        '''
        Check whether eye positions from a target are outside the fixation distance
        '''
        # Distance of an eye position from a target position
        eye_pos = self.calibrated_eye_pos
        target_pos = np.delete(self.targs[self.target_index],1)
        d_eye = np.linalg.norm(eye_pos - target_pos)
        return (d_eye > self.target_radius + self.fixation_radius_buffer) or self.pause
    
    def _start_hold(self):
        super()._start_hold()
        self.targets[self.target_index].sphere.color = target_colors[self.fixation_target_color] # change target color in fixating the target


'''
Extensions of the generic riglib.source.DataSourceSystem for getting Spikes/LFP data from the Blackrock NeuroPort system over the rig's internal network (UDP)
'''
import numpy as np
from ..source import DataSourceSystem

class Spikes(DataSourceSystem):
    '''
    For use with a DataSource in order to acquire streaming spike data from 
    the Blackrock Neural Signal Processor (NSP).
    '''

    update_freq = 30000.
    dtype = np.dtype([("ts", float), 
                      ("chan", np.int32), 
                      ("unit", np.int32),
                      ("arrival_ts", np.float64)])

    def __init__(self, channels):
        from . import cerelink
        self.conn = cerelink.Connection()
        self.conn.connect()
        self.conn.select_channels(channels)

    def start(self):
        self.conn.start_data()
        self.data = self.conn.get_event_data()

    def stop(self):
        self.conn.stop_data()

    def get(self):
        d = next(self.data)
        return np.array([(d.ts / self.update_freq, 
                          d.chan, 
                          d.unit, 
                          d.arrival_ts)],
                        dtype=self.dtype)


class LFP(DataSourceSystem):
    '''
    For use with a MultiChanDataSource in order to acquire streaming LFP 
    data from the Blackrock Neural Signal Processor (NSP).
    '''
    
    update_freq = 1000
    dtype = np.dtype('float')

    def __init__(self, channels):
        from . import cerelink
        self.conn = cerelink.Connection()
        self.conn.connect()
        self.conn.select_channels(channels)

    def start(self):
        self.conn.start_data()
        self.data = self.conn.get_continuous_data()

    def stop(self):
        self.conn.stop_data()

    def get(self):
        d = next(self.data)
        return (d.chan, d.samples)



'''Needs docs'''


import numpy as np
from OpenGL.GL import *

from .models import Model

textypes = {GL_UNSIGNED_BYTE:np.uint8, GL_FLOAT:np.float32}
class Texture(object):
    def __init__(self, tex, size=None,
        magfilter=GL_LINEAR, minfilter=GL_LINEAR, 
        wrap_x=GL_CLAMP_TO_EDGE, wrap_y=GL_CLAMP_TO_EDGE,
        iformat=GL_RGBA8, exformat=GL_RGBA, dtype=GL_UNSIGNED_BYTE):

        self.opts = dict(
            magfilter=magfilter, minfilter=minfilter, 
            wrap_x=wrap_x, wrap_y=wrap_y,
            iformat=iformat, exformat=exformat, dtype=dtype)

        if isinstance(tex, np.ndarray):
            if tex.max() <= 1:
                tex *= 255
            if len(tex.shape) < 3:
                tex = np.tile(tex, [3, 1, 1]).T
            if tex.shape[-1] == 3:
                tex = np.dstack([tex, np.ones(tex.shape[:-1])])
            size = tex.shape[:2]
            tex = tex.astype(np.uint8).tostring()
        elif isinstance(tex, str):
            im = pygame.image.load(tex)
            size = tex.get_size()
            tex = pygame.image.tostring(im, 'RGBA')
        
        self.texstr = tex
        self.size = size
        self.tex = None

    def init(self):
        gltex = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, gltex)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, self.opts['minfilter'])
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, self.opts['magfilter'])
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     self.opts['wrap_x'])
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     self.opts['wrap_y'])
        glTexImage2D(
            GL_TEXTURE_2D, 0,                           #target, level
            self.opts['iformat'],                       #internal format
            self.size[0], self.size[1], 0,              #width, height, border
            self.opts['exformat'], self.opts['dtype'],  #external format, type
            self.texstr if self.texstr is not None else 0   #pixels
        )
        
        self.tex = gltex
    
    def set(self, idx):
        glActiveTexture(GL_TEXTURE0+idx)
        glBindTexture(GL_TEXTURE_2D, self.tex)
    
    def get(self, filename=None):
        current = glGetInteger(GL_TEXTURE_BINDING_2D)
        glBindTexture(GL_TEXTURE_2D, self.tex)
        texstr = glGetTexImage(GL_TEXTURE_2D, 0, self.opts['exformat'], self.opts['dtype'])
        glBindTexture(GL_TEXTURE_2D, current)
        im = np.fromstring(texstr, dtype=textypes[self.opts['dtype']])
        im.shape = (self.size[1], self.size[0], -1)
        if filename is not None:
            np.save(filename, im)
        return im


class MultiTex(object):
    '''This is not ready yet!'''
    def __init__(self, textures, weights):
        raise NotImplementedError
        assert len(textures) < max_multitex
        self.texs = textures
        self.weights = weights

class TexModel(Model):
    def __init__(self, tex=None, **kwargs):
        if tex is not None:
            kwargs['color'] = (0,0,0,1)
        super(TexModel, self).__init__(**kwargs)
        
        self.tex = tex
    
    def init(self):
        super(TexModel, self).init()
        if self.tex.tex is None:
            self.tex.init()
        
    def render_queue(self, shader=None, **kwargs):
        if shader is not None:
            yield shader, self.draw, self.tex
        else:
            yield self.shader, self.draw, self.tex


#!/usr/bin/python
'''
Representations of plants (control systems)
'''
import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
import numpy as np
from .stereo_opengl.primitives import Cylinder, Sphere, Cone, Cube, Chain
from .bmi import robot_arms

import math
import time
import socket
import select
import numpy as np
from collections import namedtuple

from utils.constants import *

class RefTrajectories(dict):
    '''
    Generic class to hold trajectories to be replayed by a plant.
    For now, this class is just a dictionary that has had its type changed
    '''
    pass


from riglib.source import DataSourceSystem
class FeedbackData(DataSourceSystem):
    '''
    Generic class for parsing UDP feedback data from a plant. Meant to be used with
    riglib.source.DataSource to grab and log data asynchronously.

    See DataSourceSystem for notes on the source interface
    '''

    MAX_MSG_LEN = 300
    sleep_time = 0

    # must define these in subclasses
    update_freq = None
    address     = None
    dtype       = None

    def __init__(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(self.address)

        # self.file_ = open(self.feedback_filename, 'w')

    def start(self):
        self.listening = True
        self.data = self.get_feedback_data()

    def stop(self):
        self.listening = False
        self.sock.close()
        # self.file_.close()

    def __del__(self):
        # The stop commands for the socket should be issued before this object is garbage-collected, but just in case...
        self.stop()

    def get(self):
        return next(self.data)

    def get_feedback_data(self):
        '''Yield received feedback data.'''

        self.last_timestamp = -1

        while self.listening:
            r, _, _ = select.select([self.sock], [], [], 0)

            if r:  # if the list r is not empty
                feedback = self.sock.recv(self.MAX_MSG_LEN)
                ts_arrival = time.time()  # secs

                # print "feedback:", feedback
                # self.file_.write(feedback.rstrip('\r') + "\n")

                processed_feedback = self.process_received_feedback(feedback, ts_arrival)

                if processed_feedback['ts'] != self.last_timestamp:
                    yield processed_feedback

                self.last_timestamp = processed_feedback['ts']

            time.sleep(self.sleep_time)

    def process_received_feedback(self, feedback, ts_arrival):
        raise NotImplementedError('Implement in subclasses!')




class Plant(object):
    '''
    Generic interface for task-plant interaction
    '''
    hdf_attrs = []
    def __init__(self, *args, **kwargs):
        pass

    def drive(self, decoder):
        '''
        Call this function to 'drive' the plant to the state specified by the decoder

        Parameters
        ----------
        decoder : bmi.Decoder instance
            Decoder used to estimate the state of/control the plant

        Returns
        -------
        None
        '''
        # Instruct the plant to go to the decoder-specified intrinsic coordinates
        # decoder['q'] is a special __getitem__ case. See riglib.bmi.Decoder.__getitem__/__setitem__
        self.set_intrinsic_coordinates(decoder['q'])

        # Not all intrinsic coordinates will be achievable. So determine where the plant actually went
        intrinsic_coords = self.get_intrinsic_coordinates()

        # Update the decoder state with the current state of the plant, after the last command
        if not np.any(np.isnan(intrinsic_coords)):
            decoder['q'] = self.get_intrinsic_coordinates()

    def get_data_to_save(self):
        '''
        Get data to save regarding the state of the plant on every iteration of the event loop

        Parameters
        ----------
        None

        Returns
        -------
        dict:
            keys are strings, values are np.ndarray objects of data values
        '''
        return dict()

    def init(self):
        '''
        Secondary initialization after object construction. Does nothing by default
        '''
        pass

    def start(self):
        '''
        Start any auxiliary processes used by the plant
        '''
        pass

    def stop(self):
        '''
        Stop any auxiliary processes used by the plant
        '''
        pass

    def init_decoder(self, decoder):
        decoder['q'] = self.get_intrinsic_coordinates()


class AsynchronousPlant(Plant):
    def init(self):
        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()
        sink_manager.register(self.source)
        super(AsynchronousPlant, self).init()

    def start(self):
        '''
        Only start the DataSource after it has been registered with
        the sink manager in the call to init()
        '''
        self.source.start()
        super(AsynchronousPlant, self).start()

    def stop(self):
        self.source.stop()
        super(AsynchronousPlant, self).stop()

###################################################
##### Virtual plants for specific experiments #####
###################################################
class CursorPlant(Plant):
    '''
    Create a plant which is a 2-D or 3-D cursor on a screen/stereo display
    '''
    hdf_attrs = [('cursor', 'f8', (3,))]
    def __init__(self, endpt_bounds=None, cursor_radius=0.4, cursor_color=(.5, 0, .5, 1), starting_pos=np.array([0., 0., 0.]), vel_wall=True, **kwargs):
        self.endpt_bounds = endpt_bounds
        self.position = starting_pos
        self.cursor_radius = cursor_radius
        self.cursor_color = cursor_color
        self._pickle_init()
        self.vel_wall = vel_wall

    def _pickle_init(self):
        self.cursor = Sphere(radius=self.cursor_radius, color=self.cursor_color)
        self.cursor.translate(*self.position, reset=True)
        self.graphics_models = [self.cursor]

    def draw(self):
        self.cursor.translate(*self.position, reset=True)

    def set_color(self, color):
        self.cursor_color = color
        self.cursor.color = color

    def set_bounds(self, bounds):
        self.endpt_bounds = bounds

    def set_cursor_radius(self, radius):
        self.cursor_radius = radius
        self.cursor = Sphere(radius=radius, color=self.cursor_color)
        self.graphics_models = [self.cursor]

    def get_endpoint_pos(self):
        return self.position

    def set_endpoint_pos(self, pt, **kwargs):
        self.set_intrinsic_coordinates(pt)
        self.draw()

    def get_intrinsic_coordinates(self):
        return self.position

    def set_intrinsic_coordinates(self, pt):
        self.position = self._bound(pt, [])[0]
        self.draw()

    def set_visibility(self, visible):
        self.visible = visible
        if visible:
            self.graphics_models[0].attach()
        else:
            self.graphics_models[0].detach()

    def _bound(self, pos, vel):
        pos = pos.copy()
        vel = vel.copy()
        if len(vel) == 0:
            vel_wall = self.vel_wall # don't worry about vel if it's empty
            self.vel_wall = False
        if self.endpt_bounds is not None:
            if pos[0] < self.endpt_bounds[0]:
                pos[0] = self.endpt_bounds[0]
                if self.vel_wall: vel[0] = 0
            if pos[0] > self.endpt_bounds[1]:
                pos[0] = self.endpt_bounds[1]
                if self.vel_wall: vel[0] = 0

            if pos[1] < self.endpt_bounds[2]:
                pos[1] = self.endpt_bounds[2]
                if self.vel_wall: vel[1] = 0
            if pos[1] > self.endpt_bounds[3]:
                pos[1] = self.endpt_bounds[3]
                if self.vel_wall: vel[1] = 0

            if pos[2] < self.endpt_bounds[4]:
                pos[2] = self.endpt_bounds[4]
                if self.vel_wall: vel[2] = 0
            if pos[2] > self.endpt_bounds[5]:
                pos[2] = self.endpt_bounds[5]
                if self.vel_wall: vel[2] = 0
        if len(vel) == 0:
            self.vel_wall = vel_wall # restore previous value
        return pos, vel

    def drive(self, decoder):
        pos = decoder['q'].copy()
        vel = decoder['qdot'].copy()

        pos, vel = self._bound(pos, vel)

        decoder['q'] = pos
        decoder['qdot'] = vel
        super(CursorPlant, self).drive(decoder)

    def get_data_to_save(self):
        return dict(cursor=self.position)


class AuditoryCursor(Plant):
    '''
    An auditory cursor that changes frequency accordingly
    '''
    hdf_attrs = [('aud_cursor_freq', 'f8', (1,))]

    def __init__(self, min_freq, max_freq, sound_duration=0.1):
        self.min_freq = min_freq
        self.max_freq = max_freq
        self.bits = 16

        import pygame
        pygame.mixer.pre_init(44100, -self.bits, 2)
        pygame.init()

        duration = 0.1  # in seconds
        self.sample_rate = 44100
        self.n_samples = int(round(duration*self.sample_rate))
        self.max_sample = 2**(self.bits - 1) - 1
        self.freq = 0
        #setup our numpy array to handle 16 bit ints, which is what we set our mixer to expect with "bits" up above
        self.buf = np.zeros((self.n_samples, 2), dtype = np.int16)
        self.buf_ext = np.zeros((10*self.n_samples, 2), dtype=np.int16)
        self.t = np.arange(self.n_samples)/float(self.n_samples)*duration
        self.t0 = np.arange(self.n_samples)*0
        self.t_start = time.time()

    def drive(self, decoder):
        self.freq = decoder.filt.F

        if np.logical_and(decoder.cnt == 0, decoder.feedback):
            #Just got reset:
            if self.freq > self.max_freq:
                self.freq = self.max_freq
            elif self.freq < self.min_freq:
                self.freq = self.min_freq
            self.play_freq()

    def play_freq(self):
        self.buf[:,0] = np.round(self.max_sample*np.sin(2*math.pi*self.freq*self.t)).astype(int)
        self.buf[:,1] = np.round(self.max_sample*np.sin(2*math.pi*self.freq*self.t0)).astype(int)
        import pygame
        sound = pygame.sndarray.make_sound(self.buf)
        sound.play()

    def play_white_noise(self):
        self.buf_ext[:,0] = np.round(self.max_sample*np.random.normal(0, self.max_sample/2., (10*self.n_samples, ))).astype(int)
        self.buf_ext[:,1] = np.round(self.max_sample*np.zeros((10*self.n_samples, ))).astype(int)
        import pygame
        sound = pygame.sndarray.make_sound(self.buf_ext)
        sound.play()


    def get_intrinsic_coordinates(self):
        return self.freq


class onedimLFP_CursorPlant(CursorPlant):
    '''
    A square cursor confined to vertical movement
    '''
    hdf_attrs = [('lfp_cursor', 'f8', (3,))]

    def __init__(self, endpt_bounds, *args, **kwargs):
        self.lfp_cursor_rad = kwargs['lfp_cursor_rad']
        self.lfp_cursor_color = kwargs['lfp_cursor_color']
        args=[(), kwargs['lfp_cursor_color']]
        super(onedimLFP_CursorPlant, self).__init__(endpt_bounds, *args, **kwargs)


    def _pickle_init(self):
        self.cursor = Cube(side_len=self.lfp_cursor_rad, color=self.lfp_cursor_color)
        self.cursor.translate(*self.position, reset=True)
        self.graphics_models = [self.cursor]

    def drive(self, decoder):
        pos = decoder.filt.get_mean()
        pos = [-8, -2.2, pos]

        if self.endpt_bounds is not None:
            if pos[2] < self.endpt_bounds[4]:
                pos[2] = self.endpt_bounds[4]

            if pos[2] > self.endpt_bounds[5]:
                pos[2] = self.endpt_bounds[5]

            self.position = pos
            self.draw()

    def turn_off(self):
        self.cursor.detach()

    def turn_on(self):
        self.cursor.attach()

    def get_data_to_save(self):
        return dict(lfp_cursor=self.position)

class onedimLFP_CursorPlant_inverted(onedimLFP_CursorPlant):
    '''
    A square cursor confined to vertical movement
    '''
    hdf_attrs = [('lfp_cursor', 'f8', (3,))]

    def drive(self, decoder):
        std_pos = decoder.filt.get_mean()
        inv_pos = [-8, -2.2, -1.0*std_pos]

        if self.endpt_bounds is not None:
            if inv_pos[2] < self.endpt_bounds[4]:
                inv_pos[2] = self.endpt_bounds[4]

            if inv_pos[2] > self.endpt_bounds[5]:
                inv_pos[2] = self.endpt_bounds[5]

            self.position = inv_pos
            self.draw()

class twodimLFP_CursorPlant(onedimLFP_CursorPlant):
    '''Same as 1d cursor but assumes decoder returns array '''
    def drive(self, decoder):
        #Pos = (Left-Right, 0, Up-Down)
        pos = decoder.filt.get_mean()
        pos = [pos[0], -2.2, pos[2]]
        #pos = [-8, -2.2, pos[2]]

        if self.endpt_bounds is not None:
            if pos[2] < self.endpt_bounds[4]:
                pos[2] = self.endpt_bounds[4]

            if pos[2] > self.endpt_bounds[5]:
                pos[2] = self.endpt_bounds[5]

            self.position = pos
            self.draw()


arm_color = (181/256., 116/256., 96/256., 1)
arm_radius = 0.6
pi = np.pi
class RobotArmGen2D(Plant):
    '''
    Generic virtual plant for creating a kinematic chain of any number of links but confined to the X-Z (vertical) plane
    '''
    def __init__(self, link_radii=arm_radius, joint_radii=arm_radius, link_lengths=[15,15,5,5], joint_colors=arm_color,
        link_colors=arm_color, base_loc=np.array([2., 0., -15]), joint_limits=[(-pi,pi), (-pi,0), (-pi/2,pi/2), (-pi/2, 10*pi/180)], stay_on_screen=False, **kwargs):
        '''
        Instantiate the graphics and the virtual arm for a planar kinematic chain
        '''
        self.num_joints = num_joints = len(link_lengths)

        self.link_lengths = link_lengths
        self.curr_vecs = np.zeros([num_joints, 3]) #rows go from proximal to distal links

        # set initial vecs to correct orientations (arm starts out vertical)
        self.curr_vecs[0,2] = self.link_lengths[0]
        self.curr_vecs[1:,0] = self.link_lengths[1:]

        # Instantiate the kinematic chain object
        self.kin_chain = self.kin_chain_class(link_lengths, base_loc=base_loc)
        self.kin_chain.joint_limits = joint_limits

        self.base_loc = base_loc

        self.link_colors = link_colors
        self.chain = Chain(link_radii, joint_radii, link_lengths, joint_colors, link_colors)
        self.cursor = Sphere(radius=arm_radius/2, color=link_colors)
        self.graphics_models = [self.chain.link_groups[0], self.cursor]

        self.chain.translate(*self.base_loc, reset=True)

        self.hdf_attrs = [('cursor', 'f8', (3,)), ('joint_angles','f8', (self.num_joints, )), ('arm_visible', 'f8', (1,))]

        self.visible = True # arm is visible when initialized

        self.stay_on_screen = stay_on_screen
        self.joint_angles = np.zeros(self.num_joints)

    @property
    def kin_chain_class(self):
        return robot_arms.PlanarXZKinematicChain

    def set_color(self, color):
        self.link_colors = color
        self.cursor.color = color

    def set_bounds(self, bounds):
        '''
        For compatibility with other cursor plants
        '''
        pass 

    def set_cursor_radius(self, radius):
        self.cursor_radius = radius
        del self.cursor
        self.cursor = Sphere(radius=radius, color=self.link_colors)
        self.graphics_models[1] = self.cursor

    def get_endpoint_pos(self):
        '''
        Returns the current position of the non-anchored end of the arm.
        '''
        return self.kin_chain.endpoint_pos(self.joint_angles)

    def set_endpoint_pos(self, pos, **kwargs):
        '''
        Positions the arm according to specified endpoint position.
        '''
        if pos is not None:
            # Run the inverse kinematics
            angles = self.perform_ik(pos, **kwargs)

            # Update the joint configuration
            self.set_intrinsic_coordinates(angles)

    def perform_ik(self, pos, **kwargs):
        angles = self.kin_chain.inverse_kinematics(pos, q_start=self.get_intrinsic_coordinates(), verbose=False, eps=0.008, **kwargs).ravel()
        return angles

    def calc_joint_angles(self, vecs):
        return np.arctan2(vecs[:,2], vecs[:,0])

    def get_intrinsic_coordinates(self):
        '''
        Returns the joint angles of the arm in radians
        '''
        return self.calc_joint_angles(self.curr_vecs)

    def set_intrinsic_coordinates(self, theta):
        '''
        Set the joints by specifying the angles in radians.

        Parameters
        ----------
        theta : np.ndarray
            Theta is a list of angles. If an element of theta = NaN, angle should remain the same.

        Returns
        -------
        None
        '''
        new_endpt_pos = self.kin_chain.endpoint_pos(theta)
        if self.stay_on_screen and (new_endpt_pos[0] > 25 or new_endpt_pos[0] < -25 or new_endpt_pos[-1] < -14 or new_endpt_pos[-1] > 14):
            # ignore the command because it would push the endpoint off the screen
            return

        if not np.any(np.isnan(theta)):
            self.joint_angles = theta
            for i in range(self.num_joints):
                self.curr_vecs[i] = self.link_lengths[i]*np.array([np.cos(theta[i]), 0, np.sin(theta[i])])

        self.chain._update_link_graphics(self.curr_vecs)
        self.cursor.translate(*self.get_endpoint_pos(), reset=True)

    def get_data_to_save(self):
        return dict(cursor=self.get_endpoint_pos(), joint_angles=self.get_intrinsic_coordinates(), arm_visible=self.visible)

    def set_visibility(self, visible):
        self.visible = visible
        if visible:
            self.graphics_models[0].attach()
        else:
            self.graphics_models[0].detach()

class EndptControlled2LArm(RobotArmGen2D):
    '''
    2-link arm controlled in extrinsic coordinates (endpoint position)
    '''
    def __init__(self, *args, **kwargs):
        super(EndptControlled2LArm, self).__init__(*args, **kwargs)
        self.hdf_attrs = [('cursor', 'f8', (3,)), ('arm_visible','f8', (1,))]

    def get_intrinsic_coordinates(self):
        return self.get_endpoint_pos()

    def set_intrinsic_coordinates(self, pos, **kwargs):
        self.set_endpoint_pos(pos, **kwargs)

    def set_endpoint_pos(self, pos, **kwargs):
        if pos is not None:
            # Run the inverse kinematics
            theta = self.perform_ik(pos, **kwargs)
            self.joint_angles = theta

            for i in range(self.num_joints):
                if theta[i] is not None and ~np.isnan(theta[i]):
                    self.curr_vecs[i] = self.link_lengths[i]*np.array([np.cos(theta[i]), 0, np.sin(theta[i])])

            self.chain._update_link_graphics(self.curr_vecs)

    def get_data_to_save(self):
        return dict(cursor=self.get_endpoint_pos(), arm_visible=self.visible)

    @property
    def kin_chain_class(self):
        return robot_arms.PlanarXZKinematicChain2Link



'''
General-purpose "pygame" experiment class. Mostly unused
'''

import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
import pygame

from . import traits
from .experiment import LogExperiment

class Pygame(LogExperiment):
    '''
    'Window' used by the older tasks (dots, eyemove, rds, redgreen, sensorymapping)
    '''
    background = (0,0,0)
    fps = 60
    
    timeout_time = traits.Float(4., desc="Timeout (in seconds) during a trial while waiting for a response")
    penalty_time = traits.Float(5., desc="Length of penalty (in seconds) for incorrect and premature responses")
    reward_time = traits.Float(5, desc='Time of reward in seconds')

    def screen_init(self):
        '''
        Initialize the pygame display
        '''
        os.environ['SDL_VIDEO_WINDOW_POS'] = "0,0"
        os.environ['SDL_VIDEO_X11_WMCLASS'] = "monkey_experiment"

        pygame.init()
        flags = pygame.DOUBLEBUF | pygame.HWSURFACE | pygame.NOFRAME
        pygame.display.set_mode((3840,1080), flags)

        self.surf = pygame.display.get_surface()
        self.clock = pygame.time.Clock()
        self.event = None
    
    def draw_frame(self):
        raise NotImplementedError
    
    def clear_screen(self):
        self.surf.fill(self.background)
        pygame.display.flip()
    
    def _get_event(self):
        for e in pygame.event.get(pygame.KEYDOWN):
            return (e.key, e.type)
    
    def flip_wait(self):
        pygame.display.flip()
        self.event = self._get_event()
        self.clock.tick(self.fps)
    
    def _while_wait(self):
        self.surf.fill(self.background)
        self.flip_wait()
    
    def _while_trial(self):
        self.draw_frame()
        self.flip_wait()
    
    def _while_reward(self):
        self.clear_screen()
        
    def _while_penalty(self):
        self.surf.fill((181,0,45))
        self.flip_wait()
        
    def _test_start_trial(self, ts):
        return self.event is not None
    
    def _test_correct(self, ts):
        raise NotImplementedError
    
    def _test_incorrect(self, ts):
        raise NotImplementedError
    
    def _test_timeout(self, ts):
        return ts > self.timeout_time
    
    def _test_post_reward(self, ts):
        return ts > self.reward_time
    
    def _test_post_penalty(self, ts):
        return ts > self.penalty_time
    
    def _start_None(self):
        pygame.display.quit()
    
    def _start_reward(self):
        pass
    
    def _start_wait(self):
        pass



'''
BMI task features
'''
import time
import numpy as np
from riglib.experiment import traits, experiment

###### CONSTANTS
sec_per_min = 60

########################################################################################################
# Decoder/BMISystem add-ons
########################################################################################################
class NormFiringRates(traits.HasTraits):
    ''' Docstring '''
    
    norm_time = traits.Float(120., desc="Number of seconds to use for mean and SD estimate")

    def __init__(self, *args, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        raise NotImplementedError("This feature is extremely depricated and probably does not work properly anymore.")
        super(NormFiringRates, self).__init__(*args, **kwargs)
        import time
        self.starttime = time.time()
        self.elapsedtime=0
        self.count=0
        self.mFR = None
        self.mFR2 = None
        self.sdFR = None
        self.updated=False

    def update_fr_vals(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        raise NotImplementedError("decoder.bin_spikes no longer exists, use the feature extractor instead")
        if self.elapsedtime>1.:
            bin = self.decoder.bin_spikes(self.neurondata.get(all=True).copy())
            self.count +=1
            if self.count == 1:
                sz = len(bin)
                self.mFR=np.zeros([sz])
                self.mFR2=np.zeros([sz])
            delta = bin - self.mFR
            self.mFR = self.mFR + delta/self.count
            self.mFR2 = self.mFR2 + delta*(bin - self.mFR)

    def update_cursor(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.elapsedtime = time.time()-self.starttime
        if self.elapsedtime<self.norm_time:
            self.update_fr_vals()
        elif not self.updated:
            self.sdFR = np.sqrt(self.mFR2/(self.count-1))
            self.decoder.init_zscore(self.mFR,self.sdFR)
            self.hdf.sendMsg("baseline_norm")
            self.updated=True
            print("Updated session mean and SD.")
            self.hdf.sendAttr("task", "session_mFR", self.mFR)
            self.hdf.sendAttr("task", "session_sdFR", self.sdFR)

        super(NormFiringRates, self).update_cursor()

class LinearlyDecreasingAttribute(traits.HasTraits):
    ''' 
    Generic feature which linearly decreases the value of attributes used by the task
    '''
    attrs = []
    attr_flags = dict()
    def __init__(self, *args, **kwargs):
        '''
        Constructor for LinearlyDecreasingAttribute

        Parameters
        ----------
        *args, **kwargs: None necessary, these are for multiple inheritance compatibility

        Returns
        -------
        LinearlyDecreasingAttribute instance
        '''
        assert isinstance(self, experiment.Experiment)
        super(LinearlyDecreasingAttribute, self).__init__(*args, **kwargs)

    def init(self):
        '''
        Secondary init function after the object has been created and all the super __init__ functions have run.
        Initialize the "current" value of all of the attributes and register the attributes with the HDF file
        '''

        if hasattr(self, 'add_dtype'):
            for attr in self.attrs:
                if attr not in self.dtype:
                    self.add_dtype(attr, 'f8', (1,))
        
        super(LinearlyDecreasingAttribute, self).init()
        for attr in self.attrs:
            self.attr_start, self.attr_min = getattr(self, attr)
            setattr(self, 'current_%s' % attr, self.attr_start)
            self.attr_flags[attr] = True


    def _linear_change(self, start_val, end_val, decay_time):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        if start_val == end_val:
            return end_val
        else:
            elapsed_time = self.get_time() - self.task_start_time
            temp = start_val - elapsed_time/decay_time*(start_val-end_val)
            if start_val > end_val:
                return max(temp, end_val)
            elif start_val < end_val:
                return min(temp, end_val)

    def update_level(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        for attr in self.attrs:
            decay_time = float(getattr(self, '%s_time' % attr))
            attr_start, attr_min = getattr(self, attr)
            current_level = self._linear_change(attr_start, attr_min, decay_time)
            setattr(self, 'current_%s' % attr, current_level)
            flag = self.attr_flags[attr]
            if flag and getattr(self, 'current_%s' % attr) == attr_min:
                print("%s at final value after %d successful trials" % (attr, self.calc_state_occurrences('reward')))
                self.attr_flags[attr] = False

            if self.cycle_count % (1./self.update_rate * sec_per_min) == 0 and self.attr_flags[attr]:
                print("%s: " % attr, getattr(self, 'current_%s' % attr))

    def _cycle(self):
        '''
        Update and save the current attribute value before calling the next _cycle in the MRO
        '''
        self.update_level()
        if hasattr(self, 'task_data'):
            for attr in self.attrs:
                self.task_data[attr] = getattr(self, 'current_%s' % attr)

        super(LinearlyDecreasingAttribute, self)._cycle()

class LinearlyDecreasingAssist(LinearlyDecreasingAttribute):
    '''
    Specific case of LinearlyDecreasingAttribute for a linearly decreasing assist parameter
    '''
    assist_level = traits.Tuple((0.0, 0.0), desc="Level of assist to apply to BMI output")
    assist_level_time = traits.Float(600, desc="Number of seconds to go from initial to minimum assist level")

    def __init__(self, *args, **kwargs):
        super(LinearlyDecreasingAssist, self).__init__(*args, **kwargs)
        if 'assist_level' not in self.attrs:
            self.attrs.append('assist_level')

class LinearlyDecreasingXYAssist(LinearlyDecreasingAttribute):
    ''' 
    linearly decreasing XY assist -- for ArmAssist
    '''
    aa_assist_level = traits.Tuple((0.0, 0.0), desc='level of assist to apply to XY output')
    aa_assist_level_time = traits.Float(600, desc="Number of seconds to go from initial to minimum assist level")
    
    def __init__(self, *args, **kwargs):
        super(LinearlyDecreasingXYAssist, self).__init__(*args, **kwargs)
        if 'aa_assist_level' not in self.attrs:
            self.attrs.append('aa_assist_level')

class LinearlyDecreasingAngAssist(LinearlyDecreasingAttribute):
    ''' 
    linearly decreasing angular assist -- for psi and ReHand
    '''
    rh_assist_level = traits.Tuple((0.0, 0.0), desc='level of assist to apply to ang output')
    rh_assist_level_time = traits.Float(600, desc="Number of seconds to go from initial to minimum assist level")
    
    def __init__(self, *args, **kwargs):
        super(LinearlyDecreasingAngAssist, self).__init__(*args, **kwargs)
        if 'rh_assist_level' not in self.attrs:
            self.attrs.append('rh_assist_level')

class LinearlyDecreasingHalfLife(LinearlyDecreasingAttribute):
    '''
    Specific case of LinearlyDecreasingAttribute for a linearly decreasing CLDA half-life
    '''
    half_life = traits.Tuple((450., 450.), desc="Initial and final half life for CLDA")
    half_life_time = traits.Float(600, desc="Number of seconds to go from initial to final half life")
    
    def __init__(self, *args, **kwargs):
        super(LinearlyDecreasingHalfLife, self).__init__(*args, **kwargs)
        if 'half_life' not in self.attrs:
            self.attrs.append('half_life')    

class LinearlyDecreasingReachAngle(LinearlyDecreasingAssist):
    '''
    For the reach direction task, decrease the maximum reach angle linearly
    '''
    reach_angle_time = traits.Float(600, desc="Number of seconds to go from initial to final reach angle")

    def __init__(self, *args, **kwargs):
        super(LinearlyDecreasingHalfLife, self).__init__(*args, **kwargs)
        if 'half_life' not in self.attrs:
            self.attrs.append('half_life')    


import socket
import struct
import time
import numpy as np

class KinarmSocket(object):
    ''' Method for receiving UDP packets from Kinarm. The kinarm must be running the Dexterit-E task 
    called 'PK_send_UDP_kinarm', or any task with a 'UDP Send Binary' block which is located in the 
    xPC target tab of the Simulink library. Note: The normal 'UDP Send' input will NOT work -- packets
    will not be sent since the Dexterit task computer does not receive them (the xPC machine does). 

    The socket port that packets are sent to is also configured in this UDP Send Binary block. 

    Data format is double --> binary byte packing with Byte alignment = 1 (From Matlab: "The byte alignment 
        field specifies how the data types are aligned. The possible values are: 1, 2, 4, and 8. The byte 
        alignment scheme is simple, and starts each element in the list of signals on a boundary specified 
        by the alignment relative to the start of the vector."")

    Packet received are 3 x 50 matrix. Refer to Dexterit Manual (on desktop of Dexterit task computer) for 
    what the rows / columns refer to (page 71/82).  
    '''

    def __init__(self, addr=('192.168.0.8', 9090)):
        ''' Self-IP address is: 192.168.0.8 set manually. Once fully migrating system over to BMI3d, 
        will need to adjust for new IPs'''

        #Set up UDP socket (specificed by socket.SOCK_DGRAM. TCP woudl use socket.SOCK_STREAM)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) 

        #Free up port for use if you just ran task:
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        
        #Bind address
        self.socket.bind(addr)
        self.data_size = 1200 # 3 x 50 x 8 bytes (double)
        self.remote_ip = '192.168.0.2'

    def _test_connected(self):
        try:
            data, add = self.socket.recvfrom(1200)
            unpacked_ = struct.unpack('150d', data)
            assert len(unpacked_) == 150
            self.connected = True
        except:
            print('Make sure Kinarm Task is running - error in recieving packets')
            self.connected = False

    def connect(self):
        self._test_connected()

    def get_data(self):
        '''
        A generator which yields packets as they are received
        '''

        assert self.connected, "Socket is not connected, cannot get data"
        
        while self.connected:
            packet, address = self.socket.recvfrom(self.data_size)

            #Make sure packet is from correct address: 
            if address[0] == self.remote_ip:
                arrival_ts = time.time()
                data = np.array(struct.unpack('150d', packet))

                if data.shape[0] == 150:
                    #reshape data into 3 x 50
                    kindata = data.reshape(50,3).T
                    #kindata = np.hstack((kindata, np.zeros((3, 1))))
                    #kindata[:, -1] = arrival_ts
                    yield kindata

    def disconnect(self):
        self.socket.close()
        self.connected = False

    def __del__(self):
        self.disconnect()







"""Calculations of performance metrics for 2-D reaching tasks"""
import numpy as np
from scipy.stats import circmean
import matplotlib.pyplot as plt
from collections import OrderedDict, defaultdict

import os
import tables
from riglib.bmi import robot_arms, train


min_per_sec = 1./60
seconds_per_min = 60
sec_per_min = 60

pi = np.pi
plot_dir = '/storage/plots'

def _count_switches(vec):
    """ vec is an array of binary variables (0,1). The number of switches
    between 1's and 0's is counted
    """
    return len(np.nonzero(edge_detect(vec, 'pos'))[0]) + len(np.nonzero(edge_detect(vec, 'neg'))[0])

def edge_detect(vec, edge_type='pos'):
    """ Edge detector for a 1D array

    Example:

    vec = [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, ...]
                       ^           ^
                       ^           ^
                      pos         neg
                      edge        edge

    vec         : 1D array
    edge_type   : {'pos', 'neg'}
    """
    if np.ndim(vec) > 1:
        vec = vec.reshape(-1)
    T = len(vec)
    edges = np.zeros(T)
    for t in range(1,T):
        if edge_type == 'pos':
            if vec[t] and not vec[t-1]:
                edges[t] = 1
        elif edge_type == 'neg':
            if vec[t-1] and not vec[t]:
                edges[t] = 1
    return edges

def get_goal_error_basis(cursor, target):
    '''
    Define a coordinate system where one axis is along the straight-line
    from the current position to the target, and the other is orthogonal.
    Only for 2-D tasks with movement in the X-Z plane (no "depth")

    Parameters
    ----------
    cursor : np.ndarray of shape (3,)
    target : np.ndarray of shape (3,)

    Returns
    -------
    vec_to_targ : np.ndarray of shape (2,)
    orth_vec_to_targ : np.ndarray of shape (2,)
    '''
    # 90 degree rotation matrix
    R = np.array([[0, -1],
                  [1, 0]])

    vec_to_targ = target - cursor
    vec_to_targ = vec_to_targ[[0,2]]
    vec_to_targ /= np.linalg.norm(vec_to_targ)

    orth_vec_to_targ = np.dot(R, vec_to_targ)
    orth_vec_to_targ /= np.linalg.norm(orth_vec_to_targ)
    return vec_to_targ, orth_vec_to_targ

def get_task_axis_error_measures(traj, origin=np.array([0., 0.]),
    target=np.array([10., 0.]), sl=slice(None,None,1), eff_target_size=1.7-0.48,
    return_type='dict', data_rate=60., rt_from_exit_origin=False):
    '''
    calculate ME, MV, MDC, ODC, path length, and normalized path length
    for a trajectory
    '''

    if not 2 in traj.shape:
        assert "Wrong shape for trajectory!"
    elif traj.shape[1] == 2:
        traj = traj.T

    # import pdb; pdb.set_trace()

    if not np.all(origin == 0):
        traj = (traj.T - origin).T
        target -= origin
        origin = np.zeros(2)

    angle = -np.arctan2(target[1], target[0])
    R = np.array([[np.cos(angle), -np.sin(angle)],
                  [np.sin(angle), np.cos(angle)]])
    traj = np.dot(R, traj)
    ME = np.mean(np.abs(traj[1, sl]))
    MV = np.std(np.abs(traj[1, sl]))
    ODC = _count_switches( 0.5*(np.sign(np.diff(traj[0, sl])) + 1) )
    MDC = _count_switches( 0.5*(np.sign(np.diff(traj[1, sl])) + 1) )
    PL = np.sum(map(np.linalg.norm, np.diff(traj, axis=1).T))
    NPL = PL / np.linalg.norm(target - origin)

    # Determine when the cursor exits the center for the first time, i.e. ||pos - center|| > (target_radius - cursor_radius)
    dist_from_origin = np.array(map(lambda x: np.linalg.norm(x - origin), traj.T))
    try:
        if rt_from_exit_origin:
            exit_origin_ind = np.nonzero(dist_from_origin > eff_target_size)[0][0]
        else:
            exit_origin_ind = 0
        reach_time = (traj.shape[1] - exit_origin_ind)*1./data_rate#sl.step
    except:
        reach_time = np.nan

    if return_type == 'dict':
        perf = OrderedDict()

    elif return_type == 'recarray':
        data = ['ME', 'MV', 'PL', 'NPL', 'MDC', 'ODC', 'reach_time']
        dtype = np.dtype([(x, 'f8') for x in data])
        perf = np.zeros(1, dtype=dtype)
    else:
        raise ValueError("Unrecognized return type: %s" % return_type)
    perf['PL'] = PL
    perf['ME'] = ME
    perf['MV'] = MV
    perf['NPL'] = NPL
    perf['MDC'] = MDC
    perf['ODC'] = ODC
    perf['reach_time'] = reach_time
    return perf



'''
Generic features for interfacing with cortical recording systems (Plexon Omniplex, Blackrock Neuroport, etc.)
'''
from riglib import source, sink
from riglib.experiment import traits
from riglib.bmi.bmi import Decoder 
import time

class CorticalData(object):
    '''
    Feature for streaming data from Omniplex/Neuroport recording systems
    '''
    cortical_channels = None
    register_with_sink_manager = False
    send_data_to_sink_manager = False

    def init(self):
        sys_module = self.sys_module # e.g., riglib.plexon, riglib.blackrock

        kwargs = dict(send_data_to_sink_manager=self.send_data_to_sink_manager, channels=self.cortical_channels)
        if hasattr(self, "neural_src_kwargs"):
            kwargs.update(self.neural_src_kwargs)

        if hasattr(self, "_neural_src_type") and hasattr(self, "_neural_src_kwargs") and hasattr(self, "_neural_src_system_type"):
            # for testing only!
            self.neurondata = self._neural_src_type(self._neural_src_system_type, **self._neural_src_kwargs)
        elif 'spike' in self.decoder.extractor_cls.feature_type:  # e.g., 'spike_counts'
            self.neurondata = source.DataSource(sys_module.Spikes, **kwargs)
        elif 'lfp' in self.decoder.extractor_cls.feature_type:  # e.g., 'lfp_power'
            self.neurondata = source.MultiChanDataSource(sys_module.LFP, **kwargs)
        else:
            raise Exception("Unknown extractor class, unable to create data source object!")

        if self.register_with_sink_manager:
            sink_manager = sink.SinkManager.get_instance()
            sink_manager.register(self.neurondata)            

        super(CorticalData, self).init()

    @property 
    def sys_module(self):
        raise NotImplementedError("You must create a child class which specifies the recording system!")

    def run(self):
        self.neurondata.start()
        time.sleep(1) # give the datasource time to start
        try:
            super(CorticalData, self).run()
        finally:
            self.neurondata.stop()


class CorticalBMI(CorticalData, traits.HasTraits):
    '''
    Special case of CorticalData which specifies a subset of channels to stream, i.e., the ones used by the Decoder
    May not be available for all recording systems. 
    '''
    decoder = traits.InstanceFromDB(Decoder, bmi3d_db_model='Decoder', bmi3d_query_kwargs=dict())

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' sets the channels to be the channels of the decoder
        so that the PlexonData source only grabs the channels actually used by the decoder. 
        '''
        self.cortical_channels = [int(ch) for ch in self.decoder.units[:,0]]
        super(CorticalBMI, self).init()



"""
SerialLink object.

@author: Peter Corke
@copyright: Peter Corke
"""

from numpy import *
from .utility import *
from .transform import *
import copy
from .Link import *
from .plot import *

import numpy as np
from . import kinematics
from . import jacobian

def fkine(robot, q, return_allt=False):
    '''
    evaluate fkine for each point on a trajectory of
    theta_i or q_i data
    '''
    n = robot.n;
    if return_allt:
        allt = np.zeros([4, 4, n])
    else:
        allt = None

    L = robot.links;

    if len(q) == n:
        t = robot.base
        for i in range(n):
            t = t * L[i].tr(q[i])
            if return_allt:
                allt[:,:,i] = t
        t = t * robot.tool;

    else:
        raise Exception('q must have %d rows' % n)

    return t, allt
    

class SerialLink(object):
    """SerialLink object.
    Instances of this class represent a robot manipulator
    within the toolbox.
    """
        
    def __init__(self, arg=None, gravity=None, base=None, tool=None, name='', comment='', manuf=''):
        """
        SerialLink object constructor.  Create a robot from a sequence of Link objects.
        
        Several basic forms exist:
            - SerialLink()        create a null SerialLink
            - SerialLink(SerialLink)   create a clone of the SerialLink object
            - SerialLink(links)   create a SerialLink based on the passed links
            
        Various options can be set using named arguments:
        
            - gravity; gravitational acceleration (default=[0,0,9.81])
            - base; base transform (default 0)
            - tool; tool transform (default 0)
            - name
            - comment
            - manuf
        """

        if isinstance(arg, SerialLink):
            for k,v in list(arg.__dict__.items()):
                if k == "links":
                    self.__dict__[k] = copy.copy(v);           
                else:
                    self.__dict__[k] = v;           
        elif len(arg) > 1 and isinstance(arg[0], Link):
            self.links = arg;
        else:
            raise AttributeError

        # fill in default base and gravity direction
        if gravity != None:
            self.gravity = gravity;
        else:
            self.gravity = [0, 0, 9.81];
        
        if base != None:
            self.base = base;
        else:
            self.base = mat(eye(4,4));
        
        if tool != None:
            self.tool = tool;
        else:
            self.tool = mat(eye(4,4));

        if manuf:
            self.manuf = manuf
        if comment:
            self.comment = comment
        if name:
            self.name = name

        #self.handle = [];
        #self.q = [];
        #self.plotopt = {};
        #self.lineopt = {'Color', 'black', 'Linewidth', 4};
        #self.shadowopt = {'Color', 'black', 'Linewidth', 1};

        return None;

    def __str__(self):
        s = 'ROBOT(%s, %s)' % (self.name, self.config());
        return s;

    def __mul__(self, r2):
        r = SerialLink(self);        # clone the robot
        print(r)
        r.links += r2.links;
        return r;

    def copy(self):
        """
        Return a copy of the SerialLink object
        """
        return copy.copy(self);
               
    def ismdh(self):
        return self.mdh;
        
    def config(self):
        """
        Return a configuration string, one character per joint, which is
        either R for a revolute joint or P for a prismatic joint.
        For the Puma560 the string is 'RRRRRR', for the Stanford arm it is 'RRPRRR'.
        """
        s = '';
        
        for link in self.links:
            if link.sigma == 0:
                s += 'R';
            else:
                s += 'P';
        return s;

    def nofriction(self, all=False):
        """
        Return a Robot object where all friction parameters are zero.
        Useful to speed up the performance of forward dynamics calculations.
        
        @type all: boolean
        @param all: if True then also zero viscous friction
        @see: L{Link.nofriction}
        """
        r = SerialLink(self);
        r.name += "-nf";
        newlinks = [];
        for oldlink in self.links:
            newlinks.append( oldlink.nofriction(all) );
        r.links = newlinks;
        return r;
        
    def showlinks(self):
        """
        Shows details of all link parameters for this robot object, including
        inertial parameters.
        """

        count = 1;
        if self.name:
            print('name: %s'%(self.name))
        if self.manuf:
            print('manufacturer: %s'%(self.manuf))
        if self.comment:
            print('commment: %s'%(self.comment))
        for l in self.links:
            print('Link %d------------------------' % count);
            l.display()
            count += 1;
##-----------------------------------------------------------------------------------------------
##External Functions
    #From dynamics
    def accel(self, *args):
        return accel(self, *args)
    def coriolis(self, q, qd):
        return coriolis(self, q, qd)
    def inertia(self, value):
        return inertia(self,value)
    def cinertia(self, q):
        return cinertia(self, q)
    def gravload(self, q, gravity=None):
        return gravload(self, q, gravity=None)
    def rne(self, *args, **options):
        return rne(self, *args, **options)

    #from Jacobain
    def jacob0(self, q):
        return jacobian.jacob0(self,q)
    def jacobn(self, q):
        return jacobian.jacobn(self, q)

    #form Kinematics
    def fkine(self, q, **kwargs):
        return fkine(self, q, **kwargs)

    def ikine(self, tr, q=None, m=None):
        return kinematics.ikine(self, tr, q=None, m=None)
    def ikine560(robot, T, configuration=''):
        return ikine560(robot, T, configuration='')
    
    #From robot
    def fdyn(self, t0, t1, torqfun, q0, qd0, varargin):
        return fdyn(self, t0, t1, torqfun, q0, qd0, varargin)
    def itorque(self, q,qdd):
        return itorque(self, q, qdd)
    def manipulability(self, q, witch = 'yoshikawa'):
        return manipulability(self, q, witch = 'yoshikawa')
    def ospace(self, q, qd):
        return ospace(self, q, qd)

    def plot(self, q, **kwargs):
        import matplotlib.pyplot as plt
        t, allt = self.fkine(q, return_allt=True)
        joint_pos = allt[0:3,-1,:].T
        
        fig = plt.figure()
        ax = plt.subplot(111, projection='3d')
        ax.plot(joint_pos[:,0], joint_pos[:,1], joint_pos[:,2], linewidth=3)
        ax.plot(joint_pos[:,0], joint_pos[:,1], -30*np.ones_like(joint_pos[:,2]), linewidth=3, color='gray')
        
        ax.plot([0,0], [0,0], [-30, 0], linewidth=3, color='r')
        ax.set_xlim([-30, 30])
        ax.set_ylim([-30, 30])
        ax.set_zlim([-30, 30])
        
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        
        plt.show()

        #return plot(self, tg, **kwargs)
    
    
##------------------------------------------------------------------------------------------------
    def __setattr__(self, name, value):
        """
        Set attributes of the robot object
        
            - SerialLink.name = string (name of this robot)
            - SerialLink.comment = string (user comment)
            - SerialLink.manuf = string (who built it)
            - SerialLink.tool = 4x4 homogeneous tranform
            - SerialLink.base = 4x4 homogeneous tranform
            - SerialLink.gravity = 3-vector  (gx,gy,gz)
        """
        
        if name in ["manuf", "name", "comment"]:
            if not isinstance(value, str):
                raise ValueError('must be a string')
            self.__dict__[name] = value;
            
        elif name == "links":
            if not isinstance(value[0], Link):
                raise ValueError('not a Link object');
            self.__dict__[name] = value;
            self.__dict__['n'] = len(value);
            # set the robot object mdh status flag
            for link in self.links:
                if link.convention != self.links[0].convention:
                    raise Exception('robot has mixed D&H link conventions')
            self.__dict__['mdh'] = self.links[0].convention == Link.LINK_MDH;
            
        elif name == "tool":
            if not ishomog(value):
                raise ValueError('tool must be a homogeneous transform');
            self.__dict__[name] = value;

        elif name == "gravity":
            v = arg2array(value);
            if len(v) != 3:
                raise ValueError('gravity must be a 3-vector');
            self.__dict__[name] = mat(v).T
            
        elif name == "base":
            if not ishomog(value):
                raise ValueError('base must be a homogeneous transform');
            self.__dict__[name] = value;
            
        else:
            raise AttributeError;




# Generated by Django 3.1.4 on 2021-06-30 10:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0008_taskentry_metadata'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sequence',
            name='name',
            field=models.CharField(max_length=256),
        ),
    ]




'''Needs docs'''
import numpy as np
from .textures import Texture

def frustum(l, r, t, b, n, f):
    '''
    This function emulates glFrustum: https://www.opengl.org/sdk/docs/man2/xhtml/glFrustum.xml
    A frustum is a solid cut by planes, e.g., the planes representing the viewable area of a screen.

    Parameters
    ----------
    l: float
        Distance to the left plane of the screen
    r: float
        Distance to the right plane of the screen
    t: float
        Distance to the top plane of the screen
    b: float
        Distance to the bottom plane of the screen
    n: float
        Distance to the near plane of the screen
    f: float
        Distance to the far plane of the screen

    Returns
    -------
    Projection matrix to apply to solid to truncate

    '''
    rl, nrl = r + l, r - l
    tb, ntb = t + b, t - b
    fn, nfn = f + n, f - n
    return np.array([[2*n / nrl,     0,     rl / nrl,       0],
                     [    0,     2*n / ntb, tb / ntb,       0],
                     [    0,         0,    -fn / nfn, -2*f*n / nfn],
                     [    0,         0,       -1,           0]])

def perspective(angle, aspect, near, far):
    '''Generates a perspective transform matrix'''
    f = 1./ np.tan(np.radians(angle) / 2)
    fn = far + near
    nfn = far - near
    return np.array([[f/aspect, 0,    0,               0],
                     [0,        f,    0,               0],
                     [0,        0, -fn/nfn, -2*far*near/nfn],
                     [0,        0,   -1,               0]])

def orthographic(w, h, near, far):
    fn = far + near
    nfn = far - near
    return np.array([[2/w, 0,   0,      0],
                     [0,   2/h, 0,      0],
                     [0,   0,   -2/nfn, -fn/nfn],
                     [0,   0,   0,       1]])

def offaxis_frusta(winsize, fov, near, far, focal_dist, iod, flip=False, flip_z=False):
    aspect = winsize[0] / winsize[1]
    top = near * np.tan(np.radians(fov) / 2)
    right = aspect*top
    fshift = (iod/2) * near / focal_dist

    # calculate the perspective matrix for the left eye and for the right eye
    left = frustum(-right+fshift, right+fshift, top, -top, near, far)
    right = frustum(-right-fshift, right-fshift, top, -top, near, far)
    
    # multiply in the iod (intraocular distance) modelview transform
    lxfm, rxfm = np.eye(4), np.eye(4)
    lxfm[:3,-1] = [0.5*iod, 0, 0]
    rxfm[:3,-1] = [-0.5*iod, 0, 0]
    flip_mat = np.eye(4)


    if flip:
        flip_mat[0,0] = -1
    if flip_z:
        flip_mat[1,1] = -1

    return np.dot(flip_mat, np.dot(left, lxfm)), np.dot(flip_mat, np.dot(right, rxfm))

    #return np.dot(left, lxfm), np.dot(right, rxfm)

def cloudy_tex(size=(512,512)):
    '''Generates 1/f distributed noise and puts it into a texture. Looks like clouds'''
    im = np.random.randn(*size)
    grid = np.mgrid[-1:1:size[0]*1j, -1:1:size[1]*1j]
    mask = 1/(grid**2).sum(0)
    fim = np.fft.fftshift(np.fft.fft2(im))
    im = np.abs(np.fft.ifft2(np.fft.fftshift(mask * fim)))
    im -= im.min()
    return Texture(im / im.max())



'''
Base code for 'motiontracker' feature, compatible with PhaseSpace motiontracker
'''

import os
import time
import numpy as np

try:
    from OWL import *
except:
    OWL_MODE2 = False
    print("Cannot find phasespace driver")

cwd = os.path.split(os.path.abspath(__file__))[0]

class Simulate(object):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    update_freq = 240
    def __init__(self, marker_count=8, radius=(10, 2, 5), offset=(-20,0,0), speed=(5,5,4)):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.n = marker_count
        self.radius = radius
        self.offset = np.array(offset)
        self.speed = speed

        self.offsets = np.random.rand(self.n)*np.pi

    def start(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.stime = time.time()

    def get(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        time.sleep(1./self.update_freq)
        ts = (time.time() - self.stime)
        data = np.zeros((self.n, 3))
        for i, p in enumerate(self.offsets):
            x = self.radius[0] * np.cos(ts / self.speed[0] * 2*np.pi + p)
            y = self.radius[1] * np.sin(ts / self.speed[1] * 2*np.pi + p)
            z = self.radius[2] * np.sin(ts / self.speed[2] * 2*np.pi + p)
            data[i] = x,y,z

        #expands the dimension for HDFwriter saving
        data_temp = np.hstack([data + np.random.randn(self.n, 3) * 0.1, np.ones((self.n, 1))])
        data_temp_expand = np.expand_dims(data_temp, axis = 0)

        return data_temp_expand

    def stop(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        return 


class System(object):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    update_freq = 240
    def __init__(self, marker_count=8, server_name='10.0.0.11', init_flags=OWL_MODE2):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.marker_count = marker_count
        if(owlInit(server_name, init_flags) < 0):
            raise Exception(owl_get_error("init error",owlGetError()))
                
        # flush requests and check for errors fix
        if(owlGetStatus() == 0):
            raise Exception(owl_get_error("error in point tracker setup", owlGetError()))
        
        # set define frequency
        owlSetFloat(OWL_FREQUENCY, OWL_MAX_FREQUENCY)
    
        #create a point tracker
        self.tracker = 0
        owlTrackeri(self.tracker, OWL_CREATE, OWL_POINT_TRACKER)
        self._init_markers()

    def _init_markers(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        # set markers
        for i in range(self.marker_count):
            owlMarkeri(MARKER(self.tracker, i), OWL_SET_LED, i)
        owlTracker(self.tracker, OWL_ENABLE)
        self.coords = np.zeros((self.marker_count, 4))
    
    def start(self, filename=None):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.filename = filename
        if filename is not None:
            #figure out command to tell phasespace to start a recording
            pass
        owlSetInteger(OWL_STREAMING, OWL_ENABLE)
        #owlSetInteger(OWL_INTERPOLATION, 4)

    def stop(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        if self.filename is not None:
            #tell phasespace to stop recording
            pass
        owlSetInteger(OWL_STREAMING, OWL_DISABLE)
    
    def get(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        markers = []
        n = owlGetMarkers(markers, self.marker_count)
        while n == 0:
            time.sleep(.001)
            n = owlGetMarkers(markers, self.marker_count)
            
        for i, m in enumerate(markers):
            self.coords[i] = m.x, m.y, m.z, m.cond

        return self.coords

    def __del__(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        for i in range(self.marker_count):
            owlMarker(MARKER(self.tracker, i), OWL_CLEAR_MARKER)   
        owlTracker(self.tracker, OWL_DESTROY)
        owlDone()

class AligningSystem(System):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    def _init_markers(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        MAX = 32
        for i in range(self.marker_count):
            owlMarkeri(MARKER(self.tracker, i), OWL_SET_LED, i)
        for i in range(6):
            owlMarkeri(MARKER(self.tracker, self.marker_count+i), OWL_SET_LED, MAX+i)
        self.marker_count += 6
        owlTracker(self.tracker, OWL_ENABLE)
        self.coords = np.zeros((self.marker_count, 4))

def owl_get_error(s, n):
    """
    Print OWL error.
    Docstring

    Parameters
    ----------

    Returns
    -------
    """
    if(n < 0): return "%s: %d" % (s, n)
    elif(n == OWL_NO_ERROR): return "%s: No Error" % s
    elif(n == OWL_INVALID_VALUE): return "%s: Invalid Value" % s
    elif(n == OWL_INVALID_ENUM): return "%s: Invalid Enum" % s
    elif(n == OWL_INVALID_OPERATION): return "%s: Invalid Operation" % s
    else: return "%s: 0x%x" % (s, n)


def make(marker_count, cls=System, **kwargs):
    """This ridiculous function dynamically creates a class with a new init function
    Docstring

    Parameters
    ----------

    Returns
    -------
    """
    def init(self, **kwargs):
        super(self.__class__, self).__init__(marker_count=marker_count, **kwargs)
    
    dtype = np.dtype((np.float, (marker_count, 4)))
    if cls == AligningSystem:
        dtype = np.dtype((np.float, (marker_count+6, 4)))
    return type(cls.__name__, (cls,), dict(dtype=dtype, __init__=init))


def make_autoalign_reference(data, filename=os.path.join(cwd, "alignment2.npz")):
    '''Creates an alignment that can be used with the autoaligner
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    from .stereo_opengl import xfm
    assert data.shape[1:] == (6, 3)
    mdata = np.median(data,0)
    cdata = mdata - mdata[0]
    rot1 = xfm.Quaternion.rotate_vecs(np.cross(cdata[2], cdata[1]), [0,1,0])
    rdata = rot1*cdata
    rot2 = xfm.Quaternion.rotate_vecs(rdata[1], [1, 0, 0])
    np.savez(filename, data=data, reference=rot2*rot1*cdata)



'''
Virtual target capture tasks where cursors are controlled by physical
motion interfaces such as joysticks
'''
import numpy as np
from scipy.spatial.transform import Rotation as R

from riglib.experiment import traits

from .target_graphics import *
from .target_capture_task import ScreenTargetCapture, ScreenReachAngle, SequenceCapture, HandConstrainedEyeCapture, ScreenTargetCapture_Saccade
from .target_tracking_task import ScreenTargetTracking
from riglib.stereo_opengl.window import WindowDispl2D


rotations = dict(
    yzx = np.array(    # names come from rows (optitrack), but screen coords come from columns:
        [[0, 1, 0, 0], # x goes into second column (y-coordinate, coming out of screen)
        [0, 0, 1, 0],  # y goes into third column (z-coordinate, up)
        [1, 0, 0, 0],  # z goes into first column (x-coordinate, right)
        [0, 0, 0, 1]]
    ),
    zyx = np.array(
        [[0, 0, 1, 0], 
        [0, 1, 0, 0], 
        [1, 0, 0, 0], 
        [0, 0, 0, 1]]
    ),
    xzy = np.array(
        [[1, 0, 0, 0],
        [0, 0, 1, 0], 
        [0, 1, 0, 0], 
        [0, 0, 0, 1]]
    ),
    xyz = np.identity(4),
)

exp_rotations = dict(
    none = np.identity(4),
    about_x_90 = np.array(
        [[1, 0, 0, 0], 
        [0, 0, 1, 0], 
        [0, 1, 0, 0], 
        [0, 0, 0, 1]]
    ),
    about_x_minus_90 = np.array(
        [[1, 0, 0, 0], 
        [0, 0, -1, 0], 
        [0, 1, 0, 0], 
        [0, 0, 0, 1]]
    ),
    oop_xy_45 = np.array(
        [[ 0.707,  0.5  ,  0.5  , 0.],
         [ 0.   ,  0.707, -0.707, 0.],
         [-0.707,  0.5  ,  0.5  , 0.],
         [ 0.,     0.   ,  0.,    1.]]
    ),
    oop_xy_minus_45 = np.array(
        [[ 0.707,  0.5  , -0.5  , 0.],
         [ 0.   ,  0.707,  0.707, 0.],
         [ 0.707, -0.5  ,  0.5  , 0.],
         [ 0.,     0.   ,  0.,    1.]]
    ),
    oop_xy_20 = np.array(
        [[ 0.94 ,  0.117,  0.321, 0.],
         [-0.   ,  0.94 , -0.342, 0.],
         [-0.342,  0.321,  0.883, 0.],
         [ 0.,     0.   ,  0.,    1.]]
    ),
    oop_xy_minus_20 = np.array(
        [[ 0.94 ,  0.117, -0.321, 0.],
         [-0.   ,  0.94 ,  0.342, 0.],
         [ 0.342, -0.321,  0.883, 0.],
         [ 0.,     0.   ,  0.,    1.]]
    )
    
)

class ManualControlMixin(traits.HasTraits):
    '''Target capture task where the subject operates a joystick
    to control a cursor. Targets are captured by having the cursor
    dwell in the screen target for the allotted time'''

    # Settable Traits
    wait_time = traits.Float(2., desc="Time between successful trials")
    velocity_control = traits.Bool(False, desc="Position or velocity control")
    random_rewards = traits.Bool(False, desc="Add randomness to reward")
    rotation = traits.OptionsList(*rotations, desc="Control rotation matrix", bmi3d_input_options=list(rotations.keys()))
    scale = traits.Float(1.0, desc="Control scale factor")
    exp_rotation = traits.OptionsList(*exp_rotations, desc="Experimental rotation matrix", bmi3d_input_options=list(exp_rotations.keys()))
    pertubation_rotation = traits.Float(0.0, desc="rotation about bmi3d y-axis in degrees") # this is perturbation_rotation_y
    perturbation_rotation_z = traits.Float(0.0, desc="rotation about bmi3d z-axis in degrees")
    perturbation_rotation_x = traits.Float(0.0, desc="rotation about bmi3d x-axis in degrees")
    offset = traits.Array(value=[0,0,0], desc="Control offset")
    is_bmi_seed = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_pt=np.zeros([3]) #keep track of current pt
        self.last_pt = np.array(self.starting_pos) #keep track of last pt to calc. velocity
        self._quality_window_size = 500 # how many cycles to accumulate quality statistics
        self.reportstats['Input quality'] = "100 %"
        if self.random_rewards:
            self.reward_time_base = self.reward_time

    def init(self):
        self.add_dtype('manual_input', 'f8', (3,))
        super().init()
        self.no_data_counter = np.zeros((self._quality_window_size,), dtype='?')

    def _test_start_trial(self, ts):
        return ts > self.wait_time and not self.pause

    # def _test_trial_complete(self, ts):
    #     if self.target_index==self.chain_length-1 :
    #         if self.random_rewards:
    #             if not self.rand_reward_set_flag: #reward time has not been set for this iteration
    #                 self.reward_time = np.max([2*(np.random.rand()-0.5) + self.reward_time_base, self.reward_time_base/2]) #set randomly with min of base / 2
    #                 self.rand_reward_set_flag =1
    #                 #print self.reward_time, self.rand_reward_set_flag
    #         return self.target_index==self.chain_length-1

    def _test_reward_end(self, ts):
        #When finished reward, reset flag.
        if self.random_rewards:
            if ts > self.reward_time:
                self.rand_reward_set_flag = 0
                #print self.reward_time, self.rand_reward_set_flag, ts
        return ts > self.reward_time

    def _transform_coords(self, coords):
        ''' 
        Returns transformed coordinates based on rotation, offset, and scale traits
        '''
        offset = np.array(
            [[1, 0, 0, 0], 
            [0, 1, 0, 0], 
            [0, 0, 1, 0], 
            [self.offset[0], self.offset[1], self.offset[2], 1]]
        )
        scale = np.array(
            [[self.scale, 0, 0, 0], 
            [0, self.scale, 0, 0], 
            [0, 0, self.scale, 0], 
            [0, 0, 0, 1]]
        )
        old = np.concatenate((np.reshape(coords, -1), [1])) # manual input (3,) plus offset term
        new = np.linalg.multi_dot((old, offset, scale, rotations[self.rotation], exp_rotations[self.exp_rotation])) # screen coords (3,) plus offset term
        pertubation_rot = R.from_euler('y', self.pertubation_rotation, degrees=True) # this is perturb_rot_y
        perturb_rot_z = R.from_euler('z', self.perturbation_rotation_z, degrees=True)
        perturb_rot_x = R.from_euler('x', self.perturbation_rotation_x, degrees=True)
        return np.linalg.multi_dot((new[0:3], pertubation_rot.as_matrix(), perturb_rot_z.as_matrix(), perturb_rot_x.as_matrix()))

    def _get_manual_position(self):
        '''
        Fetches joystick position
        '''
        if not hasattr(self, 'joystick'):
            return
        pt = self.joystick.get()
        if len(pt) == 0:
            return

        pt = pt[-1] # Use only the latest coordinate

        if len(pt) == 2:
            pt = np.concatenate((np.reshape(pt, -1), [0]))

        return [pt]

    def move_effector(self, pos_offset=[0,0,0], vel_offset=[0,0,0]):
        ''' 
        Sets the 3D coordinates of the cursor. For manual control, uses
        motiontracker / joystick / mouse data. If no data available, returns None
        '''

        # Get raw input and save it as task data
        raw_coords = self._get_manual_position() # array of [3x1] arrays

        if raw_coords is None or len(raw_coords) < 1:
            self.no_data_counter[self.cycle_count % self._quality_window_size] = 1
            self.update_report_stats()
            self.task_data['manual_input'] = np.ones((3,))*np.nan
            return

        self.task_data['manual_input'] = raw_coords.copy()
        self.no_data_counter[self.cycle_count % self._quality_window_size] = 0

        # Transform coordinates
        coords = self._transform_coords(raw_coords)
        
        try:
            if self.limit2d:
                coords[1] = 0
            if self.limit1d:
                coords[1] = 0
                coords[0] = 0
        except:
            if self.limit2d:
                coords[1] = 0

        # Add cursor disturbance
        coords += pos_offset
        coords += vel_offset
   
        # Set cursor position
        if not self.velocity_control:
            self.current_pt = coords
        else: # TODO how to implement velocity control?
            epsilon = 2*(10**-2) # Define epsilon to stabilize cursor movement
            if sum((coords)**2) > epsilon:

                # Add the velocity (units/s) to the position (units)
                self.current_pt = coords / self.fps + self.last_pt
            else:
                self.current_pt = self.last_pt

        self.plant.set_endpoint_pos(self.current_pt)
        self.last_pt = self.plant.get_endpoint_pos()

    def update_report_stats(self):
        super().update_report_stats()
        window_size = min(max(1, self.cycle_count), self._quality_window_size)
        num_missing = np.sum(self.no_data_counter[:window_size])
        quality = 1 - num_missing / window_size
        self.reportstats['Input quality'] = "{} %".format(int(100*quality))

    @classmethod
    def get_desc(cls, params, log_summary):
        duration = round(log_summary['runtime'] / 60, 1)
        return "{}/{} succesful trials in {} min".format(
            log_summary['n_success_trials'], log_summary['n_trials'], duration)


class ManualControl(ManualControlMixin, ScreenTargetCapture):
    '''
    Slightly refactored original manual control task
    '''
    pass

class ManualControlDirectionConstraint(ManualControlMixin, ScreenReachAngle):
    '''
    Adds an additional constraint that the direction of travel must be within a certain angle
    '''
    pass

class TrackingTask(ManualControlMixin, ScreenTargetTracking):
    '''
    Track moving target task
    '''
    pass

class SequenceTask(ManualControlMixin, SequenceCapture):
    '''
    Sequence task. Subjects need to capture 3 targets. The last target appears while reaching to the 2st target.
    '''
    pass

class HandConstrainedSaccadeTask(ManualControlMixin, HandConstrainedEyeCapture):
    '''
    Saccade task while holding different targets by hand
    '''
    pass

class SaccadeTask(ManualControlMixin, ScreenTargetCapture_Saccade):
    '''
    Center out saccade task. The controller for the cursor is eye positions. The target color changes when subjects fixate the target.
    '''
    pass




'''Needs docs'''


import numpy as np
from OpenGL.GL import *

from .render import Renderer
from ..textures import Texture

fbotypes = dict(
    depth=(GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT, GL_FLOAT, GL_DEPTH_ATTACHMENT), 
    stencil=(GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, GL_STENCIL_ATTACHMENT), 
    colors=(GL_RGBA8, GL_RGBA, GL_UNSIGNED_BYTE, GL_COLOR_ATTACHMENT0)
)

class FBO(object):
    def __init__(self, attachments, size=None, ncolors=1, **kwargs):
        maxcolors = range(glGetInteger(GL_MAX_COLOR_ATTACHMENTS))
        self.names = dict(("color%d"%i, GL_COLOR_ATTACHMENT0+i) for i in maxcolors)
        self.names["depth"] = GL_DEPTH_ATTACHMENT
        self.names["stencil"] = GL_STENCIL_ATTACHMENT

        self.textures = dict()

        self.fbo = glGenFramebuffers(1)
        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        
        for attach in attachments:
            if isinstance(attach, str):
                if attach.startswith("color"):
                    idx = int(attach[5:])
                    attach = "colors"
                iform, exform, dtype, attachment = fbotypes[attach]
                texture = Texture(None, size=size, iformat=iform, exformat=exform, dtype=dtype)
                texture.init()
                if attach == "colors":
                    attachment += idx
                glFramebufferTexture2D(GL_FRAMEBUFFER, attachment, GL_TEXTURE_2D, texture.tex, 0)
                self.textures[attachment] = texture
            else:
                attachment, texture = attach
                if attachment in self.names:
                    attachment = self.names[attachment]
                if texture is None and attachment == GL_DEPTH_ATTACHMENT:
                    rb = glGenRenderbuffers(1)
                    glBindRenderbuffer(GL_RENDERBUFFER, rb)
                    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, size[0], size[1])
                    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, rb)
                else:
                    if texture.tex is None:
                        texture.init()
                    glFramebufferTexture2D(GL_FRAMEBUFFER, attachment, GL_TEXTURE_2D, texture.tex, 0)
                    self.textures[attachment] = texture
                
        types = [t for t in list(self.textures.keys()) if 
            t !=GL_DEPTH_ATTACHMENT and 
            t != GL_STENCIL_ATTACHMENT and 
            t != GL_DEPTH_STENCIL_ATTACHMENT]
        if len(types) > 0:
            glDrawBuffers(types)
        else:
            glDrawBuffers(GL_NONE)
        
        assert glCheckFramebufferStatus(GL_FRAMEBUFFER) == GL_FRAMEBUFFER_COMPLETE
        glBindFramebuffer(GL_FRAMEBUFFER, 0)
    
    def __getitem__(self, idx):
        if isinstance(idx, str):
            return self.textures[self.names[idx]]
        
        return self.textures[idx]
    
    def clear(self):
        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)
        glBindFramebuffer(GL_FRAMEBUFFER, 0)

class FBOrender(Renderer):
    def draw_fsquad(self, shader, **kwargs):
        ctx = self.programs[shader]
        glUseProgram(ctx.program)
        for name, v in list(kwargs.items()):
            if isinstance(v, Texture):
                ctx.uniforms[name] = self.get_texunit(v)
            else:
                ctx.uniforms[name] = v
        
        glEnableVertexAttribArray(ctx.attributes['position'])
        glBindBuffer(GL_ARRAY_BUFFER, self.fsquad_buf[0])
        glVertexAttribPointer(ctx.attributes['position'],
            4, GL_FLOAT, GL_FALSE, 4*4, GLvoidp(0))
        
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.fsquad_buf[1]);
        glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, GLvoidp(0))
        glDisableVertexAttribArray(ctx.attributes['position'])
    
    def draw_fsquad_to_fbo(self, fbo, shader, **kwargs):
        glBindFramebuffer(GL_FRAMEBUFFER, fbo.fbo)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)
        self.draw_fsquad(shader, **kwargs)
        glBindFramebuffer(GL_FRAMEBUFFER, 0)

    def draw_to_fbo(self, fbo, root, **kwargs):
        glBindFramebuffer(GL_FRAMEBUFFER, fbo.fbo)
        #Erase old buffer info
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)
        super(FBOrender, self).draw(root, **kwargs)
        glBindFramebuffer(GL_FRAMEBUFFER, 0)
    


'''
Reward delivery features
'''
import time
import os
import subprocess
from riglib.experiment import traits
from riglib.experiment.experiment import control_decorator
from riglib.audio import AudioPlayer
from built_in_tasks.target_graphics import VirtualRectangularTarget
import numpy as np
import serial, glob

###### CONSTANTS
sec_per_min = 60

class RewardSystem(traits.HasTraits):
    '''
    Feature for the current reward system in Amy Orsborn Lab
    '''
    trials_per_reward = traits.Float(1, desc='Number of successful trials before solenoid is opened')

    def __init__(self, *args, **kwargs):
        from riglib import reward
        super().__init__(*args, **kwargs)
        self.reward = reward.open()
        self.reportstats['Reward #'] = 0

    def run(self):
        if self.reward is None:
            raise Exception('Reward system could not be activated')
        super().run()

    def _start_reward(self):
        if hasattr(super(), '_start_reward'):
            super()._start_reward()
        self.reportstats['Reward #'] += 1
        if self.reportstats['Reward #'] % self.trials_per_reward == 0:
            self.reward.on()

    def _test_reward_end(self, ts):
        if self.reportstats['Reward #'] % self.trials_per_reward == 0:
            return ts > self.reward_time
        else:
            return True

    def _end_reward(self):
        self.reward.off()
        if hasattr(super(), '_end_reward'):
            super()._end_reward()

    @control_decorator
    def manual_reward(duration=0.5, static=True):
        from riglib import reward
        reward_sys = reward.open()
        float_dur = float(duration)  # these parameters always end up being strings
        reward_sys.async_drain(float_dur)

audio_path = os.path.join(os.path.dirname(__file__), '../riglib/audio')

class PelletReward(RewardSystem):
    '''
    Trigger pellet rewards.    
    '''
    pellets_per_reward = traits.Int(1, desc='The number of pellets to dispense per reward.')      

    def __init__(self, *args, **kwargs):
        from riglib.tablet_reward import RemoteReward
        super(RewardSystem, self).__init__(*args, **kwargs)
        self.reward = RemoteReward()
        self.reportstats['Reward #'] = 0

    def _start_reward(self):
        if hasattr(super(RewardSystem, self), '_start_reward'):
            super(RewardSystem, self)._start_reward()
        self.reportstats['Reward #'] += 1
        
        if self.reportstats['Reward #'] % self.trials_per_reward == 0:
            for _ in range(self.pellets_per_reward): # call trigger num of pellets_per_reward time
                self.reward.trigger()
                time.sleep(0.5) # wait for 0.5 seconds

    def _end_reward(self):
        if hasattr(super(RewardSystem, self), '_end_reward'):
            super(RewardSystem, self)._end_reward()

    @control_decorator
    def manual_reward( static=True):
        from riglib.tablet_reward import RemoteReward
        reward_sys = RemoteReward()
        reward_sys.reward.trigger()


class RewardAudio(traits.HasTraits):
    '''
    Play a sound in any reward state. Need to add other reward states you want to be included.
    '''

    files = [f for f in os.listdir(audio_path) if '.wav' in f]
    reward_sound = traits.OptionsList(files, desc="File in riglib/audio to play on each reward")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reward_player = AudioPlayer(self.reward_sound)

    def _start_reward(self):
        if hasattr(super(), '_start_reward'):
            super()._start_reward()
        self.reward_player.play()

class PenaltyAudio(traits.HasTraits):
    '''
    Play a sound in any penalty state. Have to define a new _start method for each different
    penalty state that might occur.
    '''
    files = list(reversed([f for f in os.listdir(audio_path) if '.wav' in f]))
    penalty_sound = traits.OptionsList(files, desc="File in riglib/audio to play on each penalty")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.penalty_player = AudioPlayer(self.penalty_sound)

    def _start_hold_penalty(self):
        if hasattr(super(), '_start_hold_penalty'):
            super()._start_hold_penalty()
        self.penalty_player.play()
    
    def _start_delay_penalty(self):
        if hasattr(super(), '_start_delay_penalty'):
            super()._start_delay_penalty()
        self.penalty_player.play()

    def _start_reach_penalty(self):
        if hasattr(super(), '_start_reach_penalty'):
            super()._start_reach_penalty()
        self.penalty_player.play()
    
    def _start_timeout_penalty(self):
        if hasattr(super(), '_start_timeout_penalty'):
            super()._start_timeout_penalty()
        self.penalty_player.play()

class PenaltyAudioMulti(traits.HasTraits):
    '''
    Separate penalty sounds for each type of penalty.
    '''
    
    hold_penalty_sound = "incorrect.wav"
    delay_penalty_sound = "buzzer.wav"
    timeout_penalty_sound = "incorrect.wav"
    reach_penalty_sound = "incorrect.wav"
    tracking_out_penalty_sound = "buzzer.wav"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hold_penalty_player = AudioPlayer(self.hold_penalty_sound)
        self.delay_penalty_player = AudioPlayer(self.delay_penalty_sound)
        self.timeout_penalty_player = AudioPlayer(self.timeout_penalty_sound)
        self.reach_penalty_player = AudioPlayer(self.reach_penalty_sound)
        self.tracking_out_penalty_player = AudioPlayer(self.tracking_out_penalty_sound)

    def _start_hold_penalty(self):
        if hasattr(super(), '_start_hold_penalty'):
            super()._start_hold_penalty()
        self.hold_penalty_player.play()
    
    def _start_delay_penalty(self):
        if hasattr(super(), '_start_delay_penalty'):
            super()._start_delay_penalty()
        self.delay_penalty_player.play()
    
    def _start_timeout_penalty(self):
        if hasattr(super(), '_start_timeout_penalty'):
            super()._start_timeout_penalty()
        self.timeout_penalty_player.play()

    def _start_reach_penalty(self):
        if hasattr(super(), '_start_reach_penalty'):
            super()._start_reach_penalty()
        self.reach_penalty_player.play()

    def _start_tracking_out_penalty(self):
        if hasattr(super(), '_start_tracking_out_penalty'):
            super()._start_tracking_out_penalty()
        self.tracking_out_penalty_player.play()

class HoldCompleteRewards(traits.HasTraits):
    '''
    Trigger an extra reward (duration set by hold_reward_time) after successful holds
    '''

    hold_reward_time = traits.Float(0.05)

    def _start_targ_transition(self):
        super()._start_targ_transition()
        if self.target_index + 1 < self.chain_length:

            # We just finished a hold/delay and there are more targets
            self.reward.async_drain(self.hold_reward_time)

class JackpotRewards(traits.HasTraits):
    '''
    Every trials_for_jackpot trials, double reward is administered
    '''

    trials_for_jackpot = traits.Int(5, desc="How many successful trials before a jackpot is delivered")

    def _test_reward_end(self, ts):
        if self.reportstats['Reward #'] % self.trials_for_jackpot == 0:
            return ts > 2*self.reward_time
        elif self.reportstats['Reward #'] % self.trials_per_reward == 0:
            return ts > self.reward_time
        else:
            return True


class ProgressBar(traits.HasTraits):
    '''
    Adds a graphical progress bar for the tracking task which fills up when the cursor is
    inside the target. Does not decrease. When the trial is over, the amount the bar filled
    up scales the amount of reward. Maximum reward is the 'reward_time' parameter.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _while_tracking_in(self):
        super()._while_tracking_in()
        
        # Update progress bar
        self.tracking_frame_index += 1
        self.tracking_rate = self.tracking_frame_index/np.shape(self.targs)[0]*self.bar_width

        if hasattr(self, 'bar'):
            for model in self.bar.graphics_models:
                self.remove_model(model)
            del self.bar

        self.bar = VirtualRectangularTarget(target_width=1.3, target_height=self.tracking_rate, target_color=(0., 1., 0., 0.75), starting_pos=[self.tracking_rate-self.bar_width,0,9])
        for model in self.bar.graphics_models:
            self.add_model(model)
        self.bar.show()

    def _while_reward(self):
        super()._while_reward()

        if hasattr(self, 'bar'):
            for model in self.bar.graphics_models:
                self.remove_model(model)
            del self.bar

        self.reward_frame_index += 1
        reward_numframe = self.reward_time*self.fps
        reward_amount = self.tracking_rate - self.reward_frame_index*self.tracking_rate/reward_numframe
        self.bar = VirtualRectangularTarget(target_width=1.3, target_height=reward_amount, target_color=(0., 1., 0., 0.75), starting_pos=[reward_amount-self.bar_width,0,9])
        for model in self.bar.graphics_models:
            self.add_model(model)
        self.bar.show()        


class TrackingRewards(traits.HasTraits):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def cleanup(self, database, saveid, **kwargs):
            self.reward.off()
            super().cleanup(database, saveid, **kwargs)

        def _start_tracking_in(self):
            super()._start_tracking_in()
            self.trigger_reward = False
            self.reward.off()
            self.reward_start_frame = self.frame_index + self.tracking_reward_interval*self.fps # frame to start first reward
            self.reward_stop_frame = self.reward_start_frame + self.tracking_reward_time*self.fps # frame to stop first reward
            # print('START TRACKING', self.reward_start_frame, self.reward_stop_frame)

        def _while_tracking_in(self):
            super()._while_tracking_in()
            # Give reward for tracking in
            if self.frame_index >= self.reward_start_frame and self.trigger_reward==False:
                self.trigger_reward = True
                self.reward_stop_frame = self.frame_index + self.tracking_reward_time*self.fps # frame to stop current reward
                self.reward_start_frame = self.frame_index + self.tracking_reward_interval*self.fps # frame to start next reward
                self.reward.on()
                # print('REWARD ON', self.frame_index/self.fps)
            if self.frame_index >= self.reward_stop_frame and self.trigger_reward==True:        
                self.trigger_reward = False
                self.reward.off()
                # print('REWARD OFF', self.frame_index/self.fps)

        def _start_tracking_out(self):
            super()._start_tracking_out()
            self.reward.off()

""""" BELOW THIS IS ALL THE OLD CODE ASSOCIATED WITH REWARD FEATURES"""


class RewardSystem_Crist(traits.HasTraits):
    '''
    Feature for the Crist solenoid reward system
    '''
    trials_per_reward = traits.Float(1, desc='Number of successful trials before solenoid is opened')

    def __init__(self, *args, **kwargs):
        from riglib import reward_crist
        super(RewardSystem, self).__init__(*args, **kwargs)
        self.reward = reward_crist.open()
        self.reportstats['Reward #'] = 0

    def _start_reward(self):
        self.reward_start = self.get_time()
        if self.reward is not None:
            self.reportstats['Reward #'] += 1
            if self.reportstats['Reward #'] % self.trials_per_reward == 0:
                self.reward.reward(self.reward_time*1000.)
        super(RewardSystem, self)._start_reward()

    def _test_reward_end(self, ts):
        if self.reportstats['Reward #'] % self.trials_per_reward == 0:
            return ts > self.reward_time
        else:
            return True

class TTLReward(traits.HasTraits):
    '''During the reward phase, send a timed TTL pulse to the reward system'''
    def __init__(self, *args, **kwargs):
        '''
        Constructor for TTLReward

        Parameters
        ----------
        pulse_device: string
            Path to the NIDAQ device used to generate the solenoid pulse
        args, kwargs: optional positional and keyword arguments to be passed to parent constructor
            None necessary

        Returns
        -------
        TTLReward instance
        '''
        import comedi
        self.com = comedi.comedi_open('/dev/comedi0')
        super(TTLReward, self).__init__(*args, **kwargs)
        self.reportstats['Reward #'] = 0

    def _start_reward(self):
        '''
        At the start of the reward state, turn on the solenoid

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        import comedi
        super(TTLReward, self)._start_reward()
        self.reportstats['Reward #'] = self.reportstats['Reward #'] + 1
        subdevice = 0
        write_mask = 0x800000
        val = 0x800000
        base_channel = 0
        comedi.comedi_dio_bitfield2(self.com, subdevice, write_mask, val, base_channel)
        self.reward_start = self.get_time() - self.start_time

    def _test_reward_end(self, ts):
        return (ts - self.reward_start) > self.reward_time

    def _end_reward(self):
        '''
        After the reward state has elapsed, turn off the solenoid

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        import comedi
        subdevice = 0
        write_mask = 0x800000
        val = 0x000000
        base_channel = 0
        comedi.comedi_dio_bitfield2(self.com, subdevice, write_mask, 0x000000, base_channel)

class TTLReward_arduino(TTLReward):
    ''' Same idea as TTL reward, using an arduino instead of nidaq (pin 8)'''
    def __init__(self, *args, **kwargs):
        self.baudrate_rew = 115200
        import serial
        self.port = serial.Serial('/dev/arduino_neurosync', baudrate=self.baudrate_rew)
        super(TTLReward_arduino, self).__init__(*args, **kwargs)
        self.reportstats['Reward #'] = 0

    def _start_reward(self):
        #port = serial.Serial('/dev/arduino_rew', baudrate=self.baudrate_rew)
        self.port.write("a")
        self.reportstats['Reward #'] = self.reportstats['Reward #'] + 1
        self.reward_start = self.get_time() - self.start_time
        super(TTLReward_arduino, self)._start_reward()
        
    def _end_reward(self):
        self.port.write("b")

class TTLReward_arduino_tdt(traits.HasTraits):
    ''' Same idea as TTL reward, using an arduino instead of nidaq (pin 8)'''
    def __init__(self, *args, **kwargs):
        self.baudrate_rew = 115200
        import serial
        #self.port = serial.Serial('/dev/ttyACM0', baudrate=self.baudrate_rew)
        self.port = serial.Serial('/dev/arduino_neurosync', baudrate=self.baudrate_rew)
        
        super(TTLReward_arduino_tdt, self).__init__(*args, **kwargs)
        self.reportstats['Reward #'] = 0

    def _start_reward(self):
        #port = serial.Serial('/dev/arduino_rew', baudrate=self.baudrate_rew)
        self.port.write("j")
        self.reportstats['Reward #'] = self.reportstats['Reward #'] + 1
        self.reward_start = self.get_time() - self.start_time
        super(TTLReward_arduino_tdt, self)._start_reward()

    def _test_reward_end(self, ts):
        return (ts - self.reward_start) > self.reward_time
        
    def _end_reward(self):
        self.port.write("n")



class JuiceLogging(traits.HasTraits):
    '''
    Save screenshots of the juice camera and link them to the task entry that has been created
    '''
    def cleanup(self, database, saveid, **kwargs):
        '''
        See riglib.experiment.Experiment.cleanup for docs on task cleanup.
        '''
        super(JuiceLogging, self).cleanup(database, saveid, **kwargs)

        ## Remove the old screenshot, if any
        fname = '/storage/temp/_juice_logging_temp.png'
        subprocess.call(['rm', fname])

        ## Use the script to run the snapshot
        subprocess.call(['camera_snapshot.sh', fname])
        # os.subprocess('camera_snapshot.sh %s' % fname)

        ## Wait for a second to make sure the file is created (TODO should really be a timed loop!)
        time.sleep(1)

        ## Link the image to the database
        dbname = kwargs['dbname'] if 'dbname' in kwargs else 'default'
        if dbname == 'default':
            database.save_data(fname, 'juice_log', saveid)
        else:
            database.save_data(fname, 'juice_log', saveid, dbname=dbname)


class ArduinoReward(traits.HasTraits):
    '''During the reward phase, send a timed TTL pulse via the Arduino microcontroller to the reward system'''
    def __init__(self, *args, **kwargs):
        '''
        Constructor for TTLReward

        Parameters
        ----------
        pulse_device: string
            Path to the NIDAQ device used to generate the solenoid pulse
        args, kwargs: optional positional and keyword arguments to be passed to parent constructor
            None necessary

        Returns
        -------
        TTLReward instance
        '''
        self.port = serial.Serial(glob.glob("/dev/ttyACM*")[0], baudrate=115200)
        #self.port.write('n')
        super(ArduinoReward, self).__init__(*args, **kwargs)
        self.reportstats['Reward #'] = 0

    def _start_reward(self):
        '''
        At the start of the reward state, turn on the solenoid

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        super(ArduinoReward, self)._start_reward()
        self.reportstats['Reward #'] = self.reportstats['Reward #'] + 1
        self.port.write('a')
        self.reward_start = self.get_time() - self.start_time

    def _test_reward_end(self, ts):
        return (ts - self.reward_start) > self.reward_time

    def _end_reward(self):
        '''
        After the reward state has elapsed, turn off the solenoid

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        self.port.write('b')


